# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions for cancelling jobs
"""

import asyncio
import json
from concurrent.futures import ThreadPoolExecutor
from typing import Any, List

from covalent._shared_files import logger
from covalent._shared_files.util_classes import RESULT_STATUS

from .. import data_manager as datasvc
from ..data_modules import job_manager
from .utils import get_executor

app_log = logger.app_log

# Dedicated thread pool for invoking non-async Executor.cancel()
_cancel_threadpool = ThreadPoolExecutor()

# Collects asyncio task futures
_background_tasks = set()


async def _cancel_task(
    dispatch_id: str, task_id: int, selected_executor: List, job_handle: str
) -> None:
    """
    Cancel the task currently being executed by the executor

    Arg(s)
        dispatch_id: Dispatch ID
        task_id: Task ID of the electron in transport graph to be cancelled
        executor: Covalent executor currently being used to execute the task
        executor_data: Executor configuration arguments
        job_handle: Unique identifier assigned to the task by the backend running the job

    Return(s)
        cancel_job_result: Status of the job cancellation action
    """
    app_log.debug(f"Cancel task {task_id} using executor {selected_executor}")
    app_log.debug(f"job_handle: {job_handle}")

    try:
        executor = get_executor(
            node_id=task_id,
            selected_executor=selected_executor,
            loop=asyncio.get_running_loop(),
            pool=_cancel_threadpool,
        )

        task_metadata = {"dispatch_id": dispatch_id, "node_id": task_id}

        cancel_job_result = await executor._cancel(task_metadata, json.loads(job_handle))
    except Exception as ex:
        app_log.debug(f"Exception when cancel task {dispatch_id}:{task_id}: {ex}")
        cancel_job_result = False

    if cancel_job_result is True:
        await job_manager.set_job_status(dispatch_id, task_id, str(RESULT_STATUS.CANCELLED))
        app_log.debug(f"Cancelled task {dispatch_id}:{task_id}")


def _to_cancel_kwargs(
    index: int, node_id: int, node_metadata: List[dict], job_metadata: List[dict]
) -> dict:
    """
    Convert node_metadata for a given node `node_id` into a dictionary

    Arg(s)
        index: Index into the node_metadata list
        node_id: Node ID
        node_metadata: List of node metadata attributes
        job_metadata: List of metadata for the current job

    Return(s)
        Node metadata dictionary
    """
    selected_executor = [node_metadata[index]["executor"], node_metadata[index]["executor_data"]]
    return {
        "task_id": node_id,
        "selected_executor": selected_executor,
        "job_handle": job_metadata[index]["job_handle"],
    }


async def cancel_tasks(dispatch_id: str, task_ids: List[int]) -> None:
    """
    Request all tasks with `task_ids` to be cancelled in the workflow identified by `dispatch_id`

    Arg(s)
        dispatch_id: Dispatch ID of the workflow
        task_ids: List of task ids to be cancelled

    Return(s)
        None
    """
    job_metadata = await job_manager.get_jobs_metadata(dispatch_id, task_ids)
    node_metadata = await _get_metadata_for_nodes(dispatch_id, task_ids)
    app_log.debug(f"node metadata: {node_metadata}")
    app_log.debug(f"job metadata: {job_metadata}")
    cancel_task_kwargs = [
        _to_cancel_kwargs(i, x, node_metadata, job_metadata) for i, x in enumerate(task_ids)
    ]

    for kwargs in cancel_task_kwargs:
        fut = asyncio.create_task(_cancel_task(dispatch_id, **kwargs))
        _background_tasks.add(fut)
        fut.add_done_callback(_background_tasks.discard)


async def _get_metadata_for_nodes(dispatch_id: str, node_ids: list) -> List[Any]:
    """
    Returns all the metadata associated with the node(s) for the workflow identified by `dispatch_id`

    Arg(s)
        dispatch_id: Dispatch ID of the workflow
        node_ids: List of node ids from the workflow to retrive the metadata for

    Return(s)
        List of node metadata for the given `node_ids`
    """

    attrs = await datasvc.electron.get_bulk(
        dispatch_id,
        node_ids,
        ["executor", "executor_data"],
    )
    return attrs



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transportable object module defining relevant classes and functions"""

import base64
import json
import platform
from typing import Any, Callable, Tuple

import cloudpickle

#  [string offset (8 bytes), big][data offset (8 bytes), big][header][string][data]

STRING_OFFSET_BYTES = 8
DATA_OFFSET_BYTES = 8
HEADER_OFFSET = STRING_OFFSET_BYTES + DATA_OFFSET_BYTES
BYTE_ORDER = "big"


class _TOArchive:

    """Archived transportable object."""

    def __init__(self, header: bytes, object_string: bytes, data: bytes):
        """
        Initialize TOArchive.

        Args:
            header: Archived transportable object header.
            object_string: Archived transportable object string.
            data: Archived transportable object data.

        Returns:
            None
        """

        self.header = header
        self.object_string = object_string
        self.data = data

    def cat(self) -> bytes:
        """
        Concatenate TOArchive.

        Returns:
            Concatenated TOArchive.

        """

        header_size = len(self.header)
        string_size = len(self.object_string)
        data_offset = STRING_OFFSET_BYTES + DATA_OFFSET_BYTES + header_size + string_size
        string_offset = STRING_OFFSET_BYTES + DATA_OFFSET_BYTES + header_size

        data_offset = data_offset.to_bytes(DATA_OFFSET_BYTES, BYTE_ORDER, signed=False)
        string_offset = string_offset.to_bytes(STRING_OFFSET_BYTES, BYTE_ORDER, signed=False)

        return string_offset + data_offset + self.header + self.object_string + self.data

    @staticmethod
    def load(serialized: bytes, header_only: bool, string_only: bool) -> "_TOArchive":
        """
        Load TOArchive object from serialized bytes.

        Args:
            serialized: Serialized transportable object.
            header_only: Load header only.
            string_only: Load string only.

        Returns:
            Archived transportable object.

        """

        string_offset = TOArchiveUtils.string_offset(serialized)
        header = TOArchiveUtils.parse_header(serialized, string_offset)
        object_string = b""
        data = b""

        if not header_only:
            data_offset = TOArchiveUtils.data_offset(serialized)
            object_string = TOArchiveUtils.parse_string(serialized, string_offset, data_offset)

            if not string_only:
                data = TOArchiveUtils.parse_data(serialized, data_offset)
        return _TOArchive(header, object_string, data)


class TOArchiveUtils:
    @staticmethod
    def data_offset(serialized: bytes) -> int:
        size64 = serialized[STRING_OFFSET_BYTES : STRING_OFFSET_BYTES + DATA_OFFSET_BYTES]
        return int.from_bytes(size64, BYTE_ORDER, signed=False)

    @staticmethod
    def string_offset(serialized: bytes) -> int:
        size64 = serialized[:STRING_OFFSET_BYTES]
        return int.from_bytes(size64, BYTE_ORDER, signed=False)

    @staticmethod
    def string_byte_range(serialized: bytes) -> Tuple[int, int]:
        """Return byte range for the string representation"""
        start_byte = TOArchiveUtils.string_offset(serialized)
        end_byte = TOArchiveUtils.data_offset(serialized)
        return start_byte, end_byte

    @staticmethod
    def data_byte_range(serialized: bytes) -> Tuple[int, int]:
        """Return byte range for the b64 picklebytes"""
        start_byte = TOArchiveUtils.data_offset(serialized)
        return start_byte, -1

    @staticmethod
    def parse_header(serialized: bytes, string_offset: int) -> bytes:
        header = serialized[HEADER_OFFSET:string_offset]
        return header

    @staticmethod
    def parse_string(serialized: bytes, string_offset: int, data_offset: int) -> bytes:
        return serialized[string_offset:data_offset]

    @staticmethod
    def parse_data(serialized: bytes, data_offset: int) -> bytes:
        return serialized[data_offset:]


class TransportableObject:
    """
    A function is converted to a transportable object by serializing it using cloudpickle
    and then whenever executing it, the transportable object is deserialized. The object
    will also contain additional info like the python version used to serialize it.

    Attributes:
        _object: The serialized object.
        python_version: The python version used on the client's machine.
    """

    def __init__(self, obj: Any) -> None:
        b64object = base64.b64encode(cloudpickle.dumps(obj))
        object_string_u8 = str(obj).encode("utf-8")

        self._object = b64object.decode("utf-8")
        self._object_string = object_string_u8.decode("utf-8")

        self._header = {
            "py_version": platform.python_version(),
            "cloudpickle_version": cloudpickle.__version__,
            "attrs": {
                "doc": getattr(obj, "__doc__", ""),
                "name": getattr(obj, "__name__", ""),
            },
        }

    @property
    def python_version(self):
        return self._header["py_version"]

    @property
    def header(self):
        return self._header

    @property
    def attrs(self):
        return self._header["attrs"]

    @property
    def object_string(self):
        # For compatibility with older Covalent
        try:
            return self._object_string
        except AttributeError:
            return self.__dict__["object_string"]

    def __eq__(self, obj) -> bool:
        if not isinstance(obj, TransportableObject):
            return False
        return self.__dict__ == obj.__dict__

    def get_deserialized(self) -> Callable:
        """
        Get the deserialized transportable object.

        Args:
            None

        Returns:
            function: The deserialized object/callable function.

        """

        return cloudpickle.loads(base64.b64decode(self._object.encode("utf-8")))

    def to_dict(self) -> dict:
        """Return a JSON-serializable dictionary representation of self"""
        return {"type": "TransportableObject", "attributes": self.__dict__.copy()}

    @staticmethod
    def from_dict(object_dict) -> "TransportableObject":
        """Rehydrate a dictionary representation

        Args:
            object_dict: a dictionary representation returned by `to_dict`

        Returns:
            A `TransportableObject` represented by `object_dict`
        """

        sc = TransportableObject(None)
        sc.__dict__ = object_dict["attributes"]
        return sc

    def get_serialized(self) -> str:
        """
        Get the serialized transportable object.

        Args:
            None

        Returns:
            object: The serialized transportable object.
        """

        return self._object

    def serialize(self) -> bytes:
        """
        Serialize the transportable object.

        Args:
            None

        Returns:
            pickled_object: The serialized object alongwith the python version.
        """

        return _to_archive(self).cat()

    def serialize_to_json(self) -> str:
        """
        Serialize the transportable object to JSON.

        Args:
            None

        Returns:
            A JSON string representation of the transportable object
        """

        return json.dumps(self.to_dict())

    @staticmethod
    def deserialize_from_json(json_string: str) -> str:
        """
        Reconstruct a transportable object from JSON

        Args:
            json_string: A JSON string representation of a TransportableObject

        Returns:
            A TransportableObject instance
        """

        object_dict = json.loads(json_string)
        return TransportableObject.from_dict(object_dict)

    @staticmethod
    def make_transportable(obj) -> "TransportableObject":
        """
        Make an object transportable.

        Args:
            obj: The object to make transportable.

        Returns:
            Transportable object.

        """

        if isinstance(obj, TransportableObject):
            return obj
        else:
            return TransportableObject(obj)

    @staticmethod
    def deserialize(
        serialized: bytes, *, header_only: bool = False, string_only: bool = False
    ) -> "TransportableObject":
        """
        Deserialize the transportable object.

        Args:
            data: serialized transportable object

        Returns:
            object: The deserialized transportable object.
        """

        ar = _TOArchive.load(serialized, header_only, string_only)
        return _from_archive(ar)

    @staticmethod
    def deserialize_list(collection: list) -> list:
        """
        Recursively deserializes a list of TransportableObjects. More
        precisely, `collection` is a list, each of whose entries is
        assumed to be either a `TransportableObject`, a list, or dict`
        """

        new_list = []
        for item in collection:
            if isinstance(item, TransportableObject):
                new_list.append(item.get_deserialized())
            elif isinstance(item, list):
                new_list.append(TransportableObject.deserialize_list(item))
            elif isinstance(item, dict):
                new_list.append(TransportableObject.deserialize_dict(item))
            else:
                raise TypeError("Couldn't deserialize collection")
        return new_list

    @staticmethod
    def deserialize_dict(collection: dict) -> dict:
        """
        Recursively deserializes a dict of TransportableObjects. More
        precisely, `collection` is a dict, each of whose entries is
        assumed to be either a `TransportableObject`, a list, or dict`

        Args:
            collection: A dictionary of TransportableObjects.
        Returns:
            A dictionary of deserialized objects.

        """

        new_dict = {}
        for k, item in collection.items():
            if isinstance(item, TransportableObject):
                new_dict[k] = item.get_deserialized()
            elif isinstance(item, list):
                new_dict[k] = TransportableObject.deserialize_list(item)
            elif isinstance(item, dict):
                new_dict[k] = TransportableObject.deserialize_dict(item)
            else:
                raise TypeError("Couldn't deserialize collection")
        return new_dict


def _to_archive(to: TransportableObject) -> _TOArchive:
    """
    Convert a TransportableObject to a _TOArchive.

    Args:
        to: Transportable object to be converted.

    Returns:
        Archived transportable object.

    """

    header = json.dumps(to._header).encode("utf-8")
    object_string = to._object_string.encode("utf-8")
    data = to._object.encode("utf-8")
    return _TOArchive(header=header, object_string=object_string, data=data)


def _from_archive(ar: _TOArchive) -> TransportableObject:
    """
    Convert a _TOArchive to a TransportableObject.

    Args:
        ar: Archived transportable object.

    Returns:
        Transportable object.

    """

    decoded_object_str = ar.object_string.decode("utf-8")
    decoded_data = ar.data.decode("utf-8")
    decoded_header = json.loads(ar.header.decode("utf-8"))
    to = TransportableObject(None)
    to._header = decoded_header
    to._object_string = decoded_object_str or ""
    to._object = decoded_data or ""

    return to



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from typing import Callable, Union

from ..._shared_files.qelectron_utils import cloudpickle_deserialize, cloudpickle_serialize
from .base import AsyncBaseQCluster, BaseQExecutor
from .default_selectors import selector_map

__all__ = [
    "QCluster",
]


class QCluster(AsyncBaseQCluster):
    """
    A cluster of quantum executors.

    Args:
        executors: A sequence of quantum executors.
        selector: A callable that selects an executor, or one of the strings "cyclic"
            or "random". The "cyclic" selector (default) cycles through `executors`
            and returns the next executor for each circuit. The "random" selector
            chooses an executor from `executors` at random for each circuit. Any
            user-defined selector must be callable with two positional arguments,
            a circuit and a list of executors. A selector must also return exactly
            one executor.
    """

    selector: Union[str, Callable] = "cyclic"

    # Flag used to indicate whether `self.selector` is currently serialized.
    # This needs to be without the "_" prefix so that it gets propagated to the server.
    selector_serialized: bool = False

    def batch_submit(self, qscripts_list):
        if self.selector_serialized:
            self.selector = self.deserialize_selector()

        selector = self.get_selector()
        selected_executor: BaseQExecutor = selector(qscripts_list, self.executors)

        # Copy server-side set attributes into selector executor.
        selected_executor.qelectron_info = self.qelectron_info.copy()
        return selected_executor.batch_submit(qscripts_list)

    def serialize_selector(self) -> None:
        if self.selector_serialized:
            return

        # serialize to bytes with cloudpickle
        self.selector = cloudpickle_serialize(self.selector)

        # convert to string to make JSON-able
        self.selector = base64.b64encode(self.selector).decode("utf-8")
        self.selector_serialized = True

    def deserialize_selector(self) -> Union[str, Callable]:
        if not self.selector_serialized:
            return self.selector

        # Deserialize the selector function (or string).
        selector = cloudpickle_deserialize(base64.b64decode(self.selector.encode("utf-8")))

        self.selector_serialized = False
        return selector

    def dict(self, *args, **kwargs) -> dict:
        # override `dict` method to convert dict attributes to JSON strings
        dict_ = super(AsyncBaseQCluster, self).dict(*args, **kwargs)
        dict_.update(executors=tuple(ex.json() for ex in self.executors))
        return dict_

    def get_selector(self) -> Callable:
        """
        Wraps `self.selector` to return defaults corresponding to string values.

        This method is called inside `batch_submit`.
        """
        self.selector = self.deserialize_selector()

        if isinstance(self.selector, str):
            # use default selector
            selector_cls = selector_map[self.selector]
            self.selector = selector_cls()

        return self.selector



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logs Routes"""

from typing import Optional

from fastapi import APIRouter, Query

from covalent_ui.api.v1.data_layer.logs_dal import Logs
from covalent_ui.api.v1.models.logs_model import SortBy, SortDirection

routes: APIRouter = APIRouter()


@routes.get("/")
def get_logs(
    count: Optional[int] = Query(0),
    offset: Optional[int] = Query(0),
    sort_by: Optional[SortBy] = SortBy.LOG_DATE,
    search: Optional[str] = "",
    sort_direction: Optional[SortDirection] = SortDirection.DESCENDING,
):
    logs = Logs()
    return logs.get_logs(sort_by, sort_direction, search, count, offset)


@routes.get("/download")
def download_logs():
    logs = Logs()
    return logs.download_logs()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logs response model"""


from typing import List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Annotated

from covalent_ui.api.v1.utils.models_helper import CaseInsensitiveEnum, SortDirection


class SortBy(CaseInsensitiveEnum):
    """Values to filter data by"""

    LOG_DATE = "log_date"
    STATUS = "status"


class LogsRequest(BaseModel):
    """Logs request model"""

    count: Annotated[int, Field(gt=0, lt=100)]
    offset: Optional[Annotated[int, Field(gt=-1)]] = 0
    sort_by: Optional[SortBy] = SortBy.LOG_DATE
    search: Optional[str] = ""
    direction: Optional[SortDirection] = SortDirection.DESCENDING


class LogsModule(BaseModel):
    """Dispatch Modeule Validation"""

    log_date: Union[str, None] = None
    status: str = "INFO"
    message: Optional[Union[str, None]] = None


class LogsResponse(BaseModel):
    """Dispatch Response Model"""

    items: List[LogsModule]
    total_count: Union[int, None] = None
    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "data": [
                    {
                        "log_date": "2022-06-13T07:45:02.114328+00:00",
                        "status": "INFO",
                        "message": "Application Started",
                    }
                ],
                "total_count": 1,
            }
        }
    )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model helper to handle sort by and sort direction"""

from enum import Enum


class CaseInsensitiveEnum(Enum):
    """Enum overriden to support case insensitive keys"""

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.upper() == value.upper():
                return member


class SortBy(CaseInsensitiveEnum):
    """Values to filter data by"""

    RUNTIME = "runtime"
    STATUS = "status"
    STARTED = "started_at"
    LATTICE_NAME = "lattice_name"
    ENDED = "ended_at"


class JobsSortBy(CaseInsensitiveEnum):
    """Values to filter jobs data by"""

    EXECUTOR = "executor"
    JOB_ID = "job_id"
    START_TIME = "start_time"
    STATUS = "status"


class SortDirection(CaseInsensitiveEnum):
    """Values to decide sort direction"""

    ASCENDING = "ASC"
    DESCENDING = "DESC"



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Functions to transform ResultSchema -> Result"""

import json
import os
from typing import Dict, Tuple

from sqlalchemy.orm import Session

from covalent._shared_files import logger
from covalent._shared_files.schemas.electron import (
    ASSET_FILENAME_MAP,
    ELECTRON_ERROR_FILENAME,
    ELECTRON_FUNCTION_FILENAME,
    ELECTRON_FUNCTION_STRING_FILENAME,
    ELECTRON_HOOKS_FILENAME,
    ELECTRON_RESULTS_FILENAME,
    ELECTRON_STDERR_FILENAME,
    ELECTRON_STDOUT_FILENAME,
    ELECTRON_STORAGE_TYPE,
    ELECTRON_VALUE_FILENAME,
    ElectronAssets,
    ElectronSchema,
)
from covalent_dispatcher._dal.asset import Asset
from covalent_dispatcher._dal.electron import ElectronMeta
from covalent_dispatcher._dal.lattice import Lattice
from covalent_dispatcher._db import models
from covalent_dispatcher._db.write_result_to_db import get_electron_type
from covalent_dispatcher._object_store.base import BaseProvider

app_log = logger.app_log


def import_electron(
    session: Session,
    dispatch_id: str,
    e: ElectronSchema,
    lat: Lattice,
    object_store: BaseProvider,
    job_id: int,
) -> Tuple[models.Electron, ElectronSchema]:
    """Returns (electron_id, ElectronSchema)"""

    electron_assets, asset_recs = import_electron_assets(
        session,
        dispatch_id,
        e,
        object_store,
    )

    # Hack for legacy DB columns
    node_storage_path = asset_recs["function"].storage_path

    electron_kwargs = _get_electron_meta(e, lat, node_storage_path, job_id)
    electron_row = ElectronMeta.create(session, insert_kwargs=electron_kwargs, flush=False)

    return (
        electron_row,
        asset_recs,
        ElectronSchema(id=e.id, metadata=e.metadata, assets=electron_assets),
    )


def _get_electron_meta(
    e: ElectronSchema, lat: Lattice, node_storage_path: str, job_id: int
) -> dict:
    kwargs = {
        "transport_graph_node_id": e.id,
        "task_group_id": e.metadata.task_group_id,
        "name": e.metadata.name,
        "executor": e.metadata.executor,
        "executor_data": json.dumps(e.metadata.executor_data),
        "qelectron_data_exists": e.metadata.qelectron_data_exists,
        "status": e.metadata.status,
        "started_at": e.metadata.start_time,
        "completed_at": e.metadata.end_time,
    }
    db_kwargs = {
        "parent_lattice_id": lat.metadata.primary_key,
        "type": get_electron_type(e.metadata.name),
        "job_id": job_id,
    }
    kwargs.update(db_kwargs)

    legacy_kwargs = {
        "storage_type": ELECTRON_STORAGE_TYPE,
        "storage_path": node_storage_path,
        "function_filename": ELECTRON_FUNCTION_FILENAME,
        "function_string_filename": ELECTRON_FUNCTION_STRING_FILENAME,
        "results_filename": ELECTRON_RESULTS_FILENAME,
        "value_filename": ELECTRON_VALUE_FILENAME,
        "stdout_filename": ELECTRON_STDOUT_FILENAME,
        "stderr_filename": ELECTRON_STDERR_FILENAME,
        "error_filename": ELECTRON_ERROR_FILENAME,
        "hooks_filename": ELECTRON_HOOKS_FILENAME,
    }
    kwargs.update(legacy_kwargs)
    return kwargs


def import_electron_assets(
    session: Session,
    dispatch_id,
    e: ElectronSchema,
    object_store: BaseProvider,
) -> Tuple[ElectronAssets, Dict[str, models.Asset]]:
    """Insert asset records


    Returns pair (ElectronAssets, asset_records), where
    `asset_records` is a mapping from asset key to asset records.

    """

    # Maps asset keys to asset records
    asset_recs = {}

    for asset_key, asset in e.assets:
        node_storage_path, object_key = object_store.get_uri_components(
            dispatch_id,
            e.id,
            asset_key,
        )

        object_key = ASSET_FILENAME_MAP[asset_key]
        local_uri = os.path.join(node_storage_path, object_key)
        asset_kwargs = {
            "storage_type": object_store.scheme,
            "storage_path": node_storage_path,
            "object_key": object_key,
            "digest_alg": asset.digest_alg,
            "digest": asset.digest,
            "remote_uri": asset.uri,
            "size": asset.size,
        }
        asset_recs[asset_key] = Asset.create(session, insert_kwargs=asset_kwargs, flush=False)

        # Send this back to the client
        asset.digest = None
        asset.remote_uri = f"file://{local_uri}"

    # Register custom assets
    if e.custom_assets:
        for asset_key, asset in e.custom_assets.items():
            object_key = f"{asset_key}.data"
            local_uri = os.path.join(node_storage_path, object_key)

            asset_kwargs = {
                "storage_type": object_store.scheme,
                "storage_path": node_storage_path,
                "object_key": object_key,
                "digest_alg": asset.digest_alg,
                "digest": asset.digest,
                "remote_uri": asset.uri,
                "size": asset.size,
            }
            asset_recs[asset_key] = Asset.create(session, insert_kwargs=asset_kwargs, flush=False)

            # Send this back to the client
            asset.remote_uri = f"file://{local_uri}" if asset.digest else ""
            asset.digest = None

    return e.assets, asset_recs



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-few-public-methods

import random

from .base import BaseQSelector


class RandomSelector(BaseQSelector):
    """
    A selector that randomly selects an executor.
    """

    name: str = "random"

    def selector_function(self, qscript, executors):
        return random.choice(executors)


class CyclicSelector(BaseQSelector):
    """
    A selector that cycles in order through the available executors.
    """

    name: str = "cyclic"

    _counter: int = 0

    def selector_function(self, qscript, executors):
        executor = executors[self._counter % len(executors)]
        self._counter += 1
        return executor


selector_map = {
    "cyclic": CyclicSelector,
    "random": RandomSelector,
}



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains all the functions required to save the decomposed result object in the database."""

import os
from datetime import datetime as dt
from datetime import timezone

import networkx as nx
from sqlalchemy import update
from sqlalchemy.orm import Session

from covalent._shared_files import logger
from covalent._shared_files.defaults import (
    arg_prefix,
    attr_prefix,
    electron_dict_prefix,
    electron_list_prefix,
    generator_prefix,
    parameter_prefix,
    prefix_separator,
    sublattice_prefix,
    subscript_prefix,
)
from covalent._shared_files.exceptions import MissingLatticeRecordError
from covalent._workflow.lattice import Lattice as LatticeClass

from .datastore import workflow_db
from .models import Asset, Electron, ElectronAsset, ElectronDependency, Job, Lattice, LatticeAsset

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class MissingElectronRecordError(Exception):
    """
    Exception to raise when an electron record is missing from the database
    """

    pass


class InvalidFileExtension(Exception):
    """
    Exception to raise when an invalid file extension is encountered
    """

    pass


def update_lattice_completed_electron_num(dispatch_id: str) -> None:
    """
    Update the number of completed electrons by one corresponding to a lattice
    """

    with workflow_db.session() as session:
        session.query(Lattice).filter_by(dispatch_id=dispatch_id).update(
            {
                "completed_electron_num": Lattice.completed_electron_num + 1,
                "updated_at": dt.now(timezone.utc),
            }
        )


def transaction_insert_lattices_data(
    session: Session,
    dispatch_id: str,
    electron_id: int,
    name: str,
    docstring_filename: str,
    electron_num: int,
    completed_electron_num: int,
    status: str,
    storage_type: str,
    storage_path: str,
    function_filename: str,
    function_string_filename: str,
    executor: str,
    executor_data: str,
    workflow_executor: str,
    workflow_executor_data: str,
    error_filename: str,
    inputs_filename: str,
    named_args_filename: str,
    named_kwargs_filename: str,
    results_filename: str,
    hooks_filename: str,
    cova_imports_filename: str,
    lattice_imports_filename: str,
    results_dir: str,
    root_dispatch_id: str,
    created_at: dt,
    updated_at: dt,
    started_at: dt,
    completed_at: dt,
) -> int:
    """
    This function writes the lattice data / metadata to the Lattices table in the DB

    Return(s)
        ID of the lattice updated
    """

    lattice_row = Lattice(
        dispatch_id=dispatch_id,
        electron_id=electron_id,
        name=name,
        docstring_filename=docstring_filename,
        status=status,
        electron_num=electron_num,
        completed_electron_num=completed_electron_num,
        storage_type=storage_type,
        storage_path=storage_path,
        function_filename=function_filename,
        function_string_filename=function_string_filename,
        executor=executor,
        executor_data=executor_data,
        workflow_executor=workflow_executor,
        workflow_executor_data=workflow_executor_data,
        error_filename=error_filename,
        inputs_filename=inputs_filename,
        named_args_filename=named_args_filename,
        named_kwargs_filename=named_kwargs_filename,
        results_filename=results_filename,
        hooks_filename=hooks_filename,
        cova_imports_filename=cova_imports_filename,
        lattice_imports_filename=lattice_imports_filename,
        results_dir=results_dir,
        root_dispatch_id=root_dispatch_id,
        is_active=True,
        created_at=created_at,
        updated_at=updated_at,
        started_at=started_at,
        completed_at=completed_at,
    )

    session.add(lattice_row)
    session.flush()
    lattice_id = lattice_row.id

    app_log.debug(f"returning lattice id {lattice_id}")
    return lattice_id


def insert_lattices_data(*args, **kwargs):
    """
    Execute the transaction to update lattice data in the database

    Return(s)
        None
    """
    with workflow_db.session() as session:
        transaction_insert_lattices_data(session, *args, **kwargs)
    app_log.debug(f"Added lattice record {locals()} to DB")


def transaction_insert_job_record(session: Session, cancel_requested: bool):
    job_row = Job(cancel_requested=cancel_requested)
    session.add(job_row)
    session.flush()
    return job_row


def transaction_insert_electron_asset_record(
    session: Session,
    electron_id: int,
    asset_id: int,
    key: str,
) -> ElectronAsset:
    electron_asset = ElectronAsset(
        meta_record_id=electron_id,
        asset_id=asset_id,
        key=key,
    )
    session.add(electron_asset)
    return electron_asset


def transaction_insert_lattice_asset_record(
    session: Session,
    lattice_id: int,
    asset_id: int,
    key: str,
) -> LatticeAsset:
    lattice_asset = LatticeAsset(
        meta_record_id=lattice_id,
        asset_id=asset_id,
        key=key,
    )
    session.add(lattice_asset)
    return lattice_asset


def transaction_insert_asset_record(
    session: Session,
    storage_type: str,
    storage_path: str,
    object_key: str,
    digest_alg: str,
    digest: str,
) -> Asset:
    asset_row = Asset(
        storage_type=storage_type,
        storage_path=storage_path,
        object_key=object_key,
        digest_alg=digest_alg,
        digest=digest,
    )
    session.add(asset_row)
    session.flush()
    return asset_row


def transaction_insert_electrons_data(
    session: Session,
    parent_dispatch_id: str,
    transport_graph_node_id: int,
    task_group_id: int,
    type: str,
    name: str,
    status: str,
    storage_type: str,
    storage_path: str,
    function_filename: str,
    function_string_filename: str,
    executor: str,
    executor_data: str,
    results_filename: str,
    value_filename: str,
    stdout_filename: str,
    stderr_filename: str,
    error_filename: str,
    hooks_filename: str,
    job_id: int,
    qelectron_data_exists: bool,
    created_at: dt,
    updated_at: dt,
    started_at: dt,
    completed_at: dt,
) -> id:
    """
    This function writes the transport graph node data to the Electrons table in the DB

    Return(s)
        electron id
    """

    # Check that the foreign key corresponding to this table exists

    row = session.query(Lattice).where(Lattice.dispatch_id == parent_dispatch_id).all()
    if len(row) == 0:
        raise MissingLatticeRecordError

    parent_lattice_id = row[0].id

    electron_row = Electron(
        parent_lattice_id=parent_lattice_id,
        transport_graph_node_id=transport_graph_node_id,
        task_group_id=task_group_id,
        type=type,
        name=name,
        status=status,
        storage_type=storage_type,
        storage_path=storage_path,
        function_filename=function_filename,
        function_string_filename=function_string_filename,
        executor=executor,
        executor_data=executor_data,
        results_filename=results_filename,
        value_filename=value_filename,
        stdout_filename=stdout_filename,
        stderr_filename=stderr_filename,
        error_filename=error_filename,
        hooks_filename=hooks_filename,
        qelectron_data_exists=qelectron_data_exists,
        is_active=True,
        job_id=job_id,
        created_at=created_at,
        updated_at=updated_at,
        started_at=started_at,
        completed_at=completed_at,
    )

    session.add(electron_row)
    session.flush()
    electron_id = electron_row.id

    return electron_id


def insert_electrons_data(*args, **kwargs):
    """
    Execute the electron update SQLalchemy transaction

    Return(s)
        electron_id
    """
    with workflow_db.session() as session:
        app_log.debug(f"Adding electron {locals()} to DB")
        return transaction_insert_electrons_data(session, *args, **kwargs)


def transaction_insert_electron_dependency_data(
    session: Session, dispatch_id: str, lattice: LatticeClass
):
    """
    Extract electron dependencies from the lattice transport graph and add them to the DB

    Return(s)
        dependency information of an electron
    """

    # TODO - Update how we access the transport graph edges directly in favor of using some interface provied by the TransportGraph class.
    node_links = nx.readwrite.node_link_data(lattice.transport_graph._graph)["links"]

    electron_dependency_ids = []

    for edge_data in node_links:
        electron_id = (
            session.query(Lattice, Electron)
            .filter(Lattice.id == Electron.parent_lattice_id)
            .filter(Lattice.dispatch_id == dispatch_id)
            .filter(Electron.transport_graph_node_id == edge_data["target"])
            .first()
            .Electron.id
        )
        parent_electron_id = (
            session.query(Lattice, Electron)
            .filter(Lattice.id == Electron.parent_lattice_id)
            .filter(Lattice.dispatch_id == dispatch_id)
            .filter(Electron.transport_graph_node_id == edge_data["source"])
            .first()
            .Electron.id
        )

        electron_dependency_row = ElectronDependency(
            electron_id=electron_id,
            parent_electron_id=parent_electron_id,
            edge_name=edge_data["edge_name"],
            parameter_type=edge_data["param_type"] if "param_type" in edge_data else None,
            arg_index=edge_data["arg_index"] if "arg_index" in edge_data else None,
            is_active=True,
            created_at=dt.now(timezone.utc),
            updated_at=dt.now(timezone.utc),
        )

        session.add(electron_dependency_row)
        electron_dependency_ids.append(electron_dependency_row.id)

    return electron_dependency_ids


def insert_electron_dependency_data(*args, **kwargs):
    """
    persist the electron dependency information into the database

    Return(s)
        None
    """
    with workflow_db.session() as session:
        app_log.debug(f"Adding electron dependency data {locals()} to DB")
        transaction_insert_electron_dependency_data(session, *args, **kwargs)


def transaction_upsert_electron_dependency_data(
    session: Session, dispatch_id: str, lattice: LatticeClass
):
    """
    Insert electron dependency records if they don't exist
    """

    electron_dependencies_exist = (
        session.query(ElectronDependency, Electron, Lattice)
        .where(
            Electron.id == ElectronDependency.electron_id,
            Electron.parent_lattice_id == Lattice.id,
            Lattice.dispatch_id == dispatch_id,
        )
        .first()
        is not None
    )
    app_log.debug(f"electron_dependencies_exist is {electron_dependencies_exist}")
    if not electron_dependencies_exist:
        transaction_insert_electron_dependency_data(
            session=session, dispatch_id=dispatch_id, lattice=lattice
        )


def transaction_update_lattices_data(session: Session, dispatch_id: str, **kwargs) -> None:
    """This function updates the lattices record."""

    valid_update = session.query(Lattice).where(Lattice.dispatch_id == dispatch_id).first()

    if not valid_update:
        raise MissingLatticeRecordError

    for attr, value in kwargs.items():
        if value:
            setattr(valid_update, attr, value)

    session.add(valid_update)


def update_lattices_data(dispatch_id: str, **kwargs) -> None:
    """This function updates the lattices record."""

    with workflow_db.session() as session:
        transaction_update_lattices_data(session, dispatch_id, **kwargs)


def update_electrons_data(
    parent_dispatch_id: str,
    transport_graph_node_id: int,
    name: str,
    status: str,
    started_at: dt,
    updated_at: dt,
    completed_at: dt,
    qelectron_data_exists: bool,
) -> None:
    """This function updates the electrons record."""

    with workflow_db.session() as session:
        parent_lattice_id = (
            session.query(Lattice).where(Lattice.dispatch_id == parent_dispatch_id).all()[0].id
        )
        valid_update = (
            session.query(Electron)
            .where(
                Electron.parent_lattice_id == parent_lattice_id,
                Electron.transport_graph_node_id == transport_graph_node_id,
            )
            .first()
            is not None
        )
        if not valid_update:
            raise MissingElectronRecordError

        session.execute(
            update(Electron)
            .where(
                Electron.parent_lattice_id == parent_lattice_id,
                Electron.transport_graph_node_id == transport_graph_node_id,
            )
            .values(
                name=name,
                status=status,
                started_at=started_at,
                updated_at=updated_at,
                completed_at=completed_at,
                qelectron_data_exists=qelectron_data_exists,
            )
        )


def get_electron_type(node_name: str) -> str:
    """Get the electron type (to be written to DB) given the electron node data."""

    if node_name.startswith(arg_prefix):
        return arg_prefix.strip(prefix_separator)

    elif node_name.startswith(attr_prefix):
        return attr_prefix.strip(prefix_separator)

    elif node_name.startswith(electron_dict_prefix):
        return electron_dict_prefix.strip(prefix_separator)

    elif node_name.startswith(electron_list_prefix):
        return electron_list_prefix.strip(prefix_separator)

    elif node_name.startswith(generator_prefix):
        return generator_prefix.strip(prefix_separator)

    elif node_name.startswith(parameter_prefix):
        return parameter_prefix.strip(prefix_separator)

    elif node_name.startswith(sublattice_prefix):
        return sublattice_prefix.strip(prefix_separator)

    elif node_name.startswith(subscript_prefix):
        return subscript_prefix.strip(prefix_separator)

    else:
        return "function"


def get_sublattice_electron_id(parent_dispatch_id: str, sublattice_node_id: int):
    """
    Query the electron ID is a sublattice

    Return(s)
        sublattice electron id
    """
    with workflow_db.session() as session:
        sublattice_electron_id = (
            session.query(Lattice, Electron)
            .where(
                Lattice.id == Electron.parent_lattice_id,
                Electron.transport_graph_node_id == sublattice_node_id,
                Lattice.dispatch_id == parent_dispatch_id,
            )
            .first()
            .Electron.id
        )

    return sublattice_electron_id


def resolve_electron_id(eid: int):
    """
    Given an electron's unique id, return the corresponding dispatch_id and node_id

    Return(s)
        dispatch_id of the lattice
        ID of the node within the lattice
    """
    with workflow_db.session() as session:
        row = (
            session.query(Lattice, Electron)
            .where(Electron.id == eid, Lattice.id == Electron.parent_lattice_id)
            .first()
        )
        dispatch_id = row.Lattice.dispatch_id
        node_id = row.Electron.transport_graph_node_id

    return dispatch_id, node_id


def write_lattice_error(dispatch_id: str, error: str) -> None:
    """
    Persist the lattice error into the database

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        error: Lattice error to be persisted

    Return(s)
        None
    """
    with workflow_db.session() as session:
        valid_update = session.query(Lattice).where(Lattice.dispatch_id == dispatch_id).first()

        if not valid_update:
            raise MissingLatticeRecordError

        with open(os.path.join(valid_update.storage_path, valid_update.error_filename), "w") as f:
            f.write(error)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from pathlib import Path
from typing import Callable, Tuple

from furl import furl

from ..._shared_files import logger
from .. import File
from .transfer_strategy_base import FileTransferStrategy

app_log = logger.app_log


class GCloud(FileTransferStrategy):
    """Implements FileTransferStrategy class to transfer files to/from Google Cloud Storage.

    Args:
        credentials: Path to OAuth 2.0 credentials JSON file for a service account
        project_id: ID of a project in GCP

    Attributes:
        credentials: String containing OAuth 2.0 credentials
        project_id: ID of a project in GCP
    """

    def __init__(self, credentials: str = None, project_id: str = None):
        if credentials is not None:
            credentials_json = Path(credentials).expanduser().resolve()

            if not credentials_json.is_file():
                raise ValueError("Cloud not locate credentials for GCloud file transfers.")

            with open(credentials_json) as f:
                self.credentials = json.load(f)
        else:
            self.credentials = None

        self.project_id = project_id

    def _get_service_client(self, bucket_name: str):
        """Returns the service client object for the Blob storage account.

        Args:
            bucket_name: Name of a storage bucket in Google Cloud

        Returns:
            Storage client object for the Google Cloud storage service
        """

        try:
            from google.cloud import storage
            from google.oauth2 import service_account
        except ImportError:
            print("Missing required dependency. Reinstall package as `pip install covalent[gcp]`.")
            raise

        return storage.Client(
            project=self.project_id,
            credentials=service_account.Credentials.from_service_account_info(self.credentials),
        ).bucket(bucket_name)

    def _parse_uri(self, gcloud_uri: str) -> Tuple[str, str]:
        """Parses a GCloud object URI and returns the bucket name and the object path.

        Args:
            blob_uri: A URI for an Azure Blob object in the form https://<storage_account_name>.blob.core.windows.net/<container_name>/<blob_name>

        Returns:
            parsed_uri: A tuple containing the storage account name, container name, and blob name
        """

        object_furl = furl(gcloud_uri)

        bucket_name = object_furl.host
        object_path = object_furl.path

        if str(object_path).startswith("/"):
            object_path = Path(str(object_path)[1:])

        return (bucket_name, object_path)

    def _download_file(self, blob_client, destination_path: str) -> None:
        blob_client.download_to_filename(destination_path)

    def download(self, from_file: File, to_file: File = File()) -> Callable:
        from_filepath = from_file.filepath
        to_filepath = to_file.filepath

        bucket_name, object_path = self._parse_uri(from_file.uri)

        app_log.debug(
            f"GCloud download; bucket name: {bucket_name}, from_filepath: {from_filepath}, to_filepath {to_filepath}."
        )

        def callable():
            """Download file or directory from a Google Cloud Storage bucket."""
            from pathlib import Path

            service_client = self._get_service_client(bucket_name)

            blobs = service_client.list_blobs(prefix=object_path)
            for blob in blobs:
                if blob.name.endswith("/"):
                    continue

                dest_obj_path = (
                    Path(to_filepath) / Path(blob.name).relative_to(object_path)
                    if to_file.is_dir
                    else Path(to_filepath)
                )
                dest_obj_path.parent.mkdir(parents=True, exist_ok=True)
                self._download_file(blob, str(dest_obj_path))

        return callable

    def _upload_file(self, service_client, source_path: str, destination_path: Path) -> None:
        blob = service_client.blob(destination_path)
        blob.upload_from_filename(source_path)

    def upload(self, from_file: File, to_file: File = File()) -> Callable:
        from_filepath = from_file.filepath
        to_filepath = to_file.filepath

        bucket_name, object_path = self._parse_uri(to_file.uri)

        app_log.debug(
            f"GCloud upload; bucket name: {bucket_name}, from_filepath: {from_filepath}, to_filepath {to_filepath}."
        )

        def callable():
            """Upload file or directory to a Google Cloud Storage bucket."""
            from pathlib import Path

            service_client = self._get_service_client(bucket_name)

            files = (
                [path for path in Path(from_filepath).rglob("*") if path.is_file()]
                if from_file.is_dir
                else [from_filepath]
            )
            for file in files:
                dest_obj_path = (
                    Path(object_path) / Path(file).relative_to(from_filepath)
                    if to_file.is_dir
                    else Path(object_path)
                )
                self._upload_file(service_client, file, str(dest_obj_path))

        return callable

    def cp(self, from_file: File, to_file: File = File()) -> File:  # pragma: no cover
        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Electrons Schema """

from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Text, func

from covalent_ui.api.v1.database.config.db import Base


class Electron(Base):
    """Electron

    Attributes:
        id: primary key id
        parent_lattice_id: id of the lattice containing this electron
        transport_graph_node_id: id of the node in the context of a transport graph
        task_group_id: id of the node's task group in the context of a transport graph
        type: node type
        name: node name
        status: Execution status of the node
        storage_type: Storage backend type for data files ("local", "s3")
        storage_path: Bucket name (dispatch_id)
        function_filename: Name of the file containing the serialized function
        function_string_filename: Name of the file containing the function string
        executor_filename: Name of the file containing the serialized executor
        error_filename: Name of the file containing an error message for the electron
        results_filename: Name of the file containing the serialized output
        value_filename: Name of the file containing the serialized parameter (for parameter nodes)
        attribute_name: Electron attribute name (for attribute nodes)
        key: For generated and subscript nodes
        stdout_filename: Name of the file containing standard output generated by the task
        stderr_filename: Name of the file containing standard error generated by the task
        hooks_filename: Name of the file containing depends instances of DepsBash, DepsPip, and DepsCall
        error_filename: Name of the file containing execution information generated at runtime
        is_active: Status of the record, 1: active and 0: inactive
        job_id: ID for circuit_info
        qelectron_data_exists: Flag that indicates if qelectron data exists in the electron
        created_at: created timestamp
        updated_at: updated timestamp
        started_at: started timestamp
        completed_at: completed timestamp
    """

    __tablename__ = "electrons"
    id = Column(Integer, primary_key=True)

    # id of the lattice containing this electron
    parent_lattice_id = Column(Integer, ForeignKey("lattices.id"), nullable=False)

    # id of the node in the context of a transport graph
    transport_graph_node_id = Column(Integer, nullable=False)

    # id of the node's task group in the context of a transport graph
    task_group_id = Column(Integer, nullable=False)

    # Node type
    type = Column(String(24), nullable=False)

    # Node name
    name = Column(Text, nullable=False)

    # Execution status of the node
    status = Column(String(24), nullable=False)

    # Storage backend type for data files ("local", "s3")
    storage_type = Column(Text)

    # Bucket name (dispatch_id)
    storage_path = Column(Text)

    # Name of the file containing the serialized function
    function_filename = Column(Text)

    # Name of the file containing the function string
    function_string_filename = Column(Text)

    # Short name describing the executor ("local", "dask", etc)
    executor = Column(Text)

    # JSONified executor attributes
    executor_data = Column(Text)

    # name of the file containing the serialized output
    results_filename = Column(Text)

    # Name of the file containing the serialized parameter (for parameter nodes)
    value_filename = Column(Text)

    # Name of the file containing standard output generated by the task
    stdout_filename = Column(Text)

    # Name of the file containing the electron execution dependencies
    hooks_filename = Column(Text)

    # Name of the file containing standard error generated by the task
    stderr_filename = Column(Text)

    # Name of the file containing errors generated by the task runner or executor
    error_filename = Column(Text)

    # Name of the column which signifies soft deletion of the electrons corresponding to a lattice
    is_active = Column(Boolean, nullable=False, default=True)

    # ID for circuit_info
    job_id = Column(Integer, ForeignKey("jobs.id", name="job_id_link"), nullable=False)

    # Flag that indicates if qelectron data exists in the electron
    qelectron_data_exists = Column(Boolean, nullable=False, default=False)

    # Timestamps
    created_at = Column(DateTime, nullable=False, server_default=func.now())
    updated_at = Column(DateTime, nullable=False, onupdate=func.now(), server_default=func.now())
    started_at = Column(DateTime)
    completed_at = Column(DateTime)


class Job(Base):
    __tablename__ = "jobs"
    id = Column(Integer, primary_key=True)

    # Indicates whether the job has been requested to be cancelled
    cancel_requested = Column(Boolean, nullable=False, default=False)

    # Indicates whether the task cancellation succeeded (return value
    # of Executor.cancel())
    cancel_successful = Column(Boolean, nullable=False, default=False)

    # JSON-serialized identifier for job
    job_handle = Column(Text, nullable=False, default="null")



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mappings between electron attributes and DB records"""

import json

from covalent._shared_files.schemas.electron import ELECTRON_ASSET_KEYS, ELECTRON_METADATA_KEYS
from covalent._shared_files.util_classes import Status

METADATA_KEYS = ELECTRON_METADATA_KEYS
ASSET_KEYS = ELECTRON_ASSET_KEYS

COMPUTED_FIELDS = {"sub_dispatch_id"}

_meta_record_map = {
    "node_id": "transport_graph_node_id",
    "task_group_id": "task_group_id",
    "name": "name",
    "start_time": "started_at",
    "end_time": "completed_at",
    "status": "status",
    "executor": "executor",
    "executor_data": "executor_data",
    "qelectron_data_exists": "qelectron_data_exists",
}

_db_meta_record_map = {
    "id": "id",
    "parent_lattice_id": "parent_lattice_id",
    "type": "type",
    "storage_path": "storage_path",
    "storage_type": "storage_type",
    "job_id": "job_id",
}

_meta_record_map.update(_db_meta_record_map)


def identity(x):
    return x


def get_status_filter(raw: str):
    return Status(raw)


def set_status_filter(stat: Status):
    return str(stat)


def get_executor_data_filter(raw: str):
    return json.loads(raw)


def set_executor_data_filter(object_dict: dict):
    return json.dumps(object_dict)


custom_get_filters = {
    "status": get_status_filter,
    "executor_data": get_executor_data_filter,
    "type": identity,
    "sub_dispatch_id": identity,
}

custom_set_filters = {"status": set_status_filter, "executor_data": set_executor_data_filter}

get_filters = {key: identity for key in METADATA_KEYS.union(ASSET_KEYS)}
set_filters = {key: identity for key in METADATA_KEYS.union(ASSET_KEYS)}

get_filters.update(custom_get_filters)
set_filters.update(custom_set_filters)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
from datetime import datetime

from fastapi.responses import Response

from covalent._shared_files.config import get_config

UI_LOGFILE = get_config("user_interface.log_dir") + "/covalent_ui.log"


class Logs:
    """Logs data access layer"""

    def __init__(self) -> None:
        self.config = get_config

    def get_logs(self, sort_by, direction, search, count, offset):
        with open(UI_LOGFILE, "r", encoding="utf-8") as logfile:
            search.lower()
            unmatch_str = ""
            log = []
            reverse_list = direction.value == "DESC"
            for i in logfile:
                split_reg = r"\[(.*)\] \[(TRACE|DEBUG|INFO|NOTICE|WARN|WARNING|ERROR|SEVERE|CRITICAL|FATAL)\]"  # r"\[(.*)\] \[(.*)\] ((.|\n)*)"
                data = re.split(pattern=split_reg, string=i)
                if len(data) > 1:
                    try:
                        parse_str = datetime.strptime(data[1], "%Y-%m-%d %H:%M:%S,%f")
                        json_data = {
                            "log_date": f"{parse_str}",
                            "status": data[2],
                            "message": data[3],
                        }
                    except ValueError:
                        json_data = {"log_date": f"{None}", "status": data[2], "message": data[1:]}
                    log.append(json_data)
                else:
                    len_log = len(log)
                    unmatch_str += i
                    if len_log > 0 and ((unmatch_str != "")):
                        msg = log[len_log - 1]["message"] + "\n"
                        log[len_log - 1]["message"] = msg + unmatch_str
                        unmatch_str = ""
                    else:
                        log.append({"log_date": None, "status": "INFO", "message": unmatch_str})
                        unmatch_str = ""
            log = [
                i
                for i in log
                if (i["message"].lower().__contains__(search))
                or (i["status"].lower().__contains__(search))
            ]
            result = sorted(
                log,
                key=lambda e: (e[sort_by.value] is not None, e[sort_by.value]),
                reverse=reverse_list,
            )
            total_count = len(result)
            result = result[offset : count + offset] if count != 0 else log[offset:]
            return {"items": result, "total_count": total_count}

    def download_logs(self):
        """Download logs"""
        data = None
        if os.path.exists(UI_LOGFILE):
            with open(UI_LOGFILE, "rb") as file:
                data = file.read().decode("utf-8")
                return Response(data)
        return {"data": data}



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Serialization/Deserialization methods for Assets """

import hashlib
import json
from enum import Enum
from pathlib import Path
from typing import Any

import cloudpickle

from .._shared_files.schemas.asset import AssetSchema
from .._workflow.transportable_object import TransportableObject

__all__ = [
    "AssetType",
    "save_asset",
    "load_asset",
]


CHECKSUM_ALGORITHM = "sha"


class AssetType(Enum):
    """
    Enum for the type of Asset data

    """

    OBJECT = 0  # Fallback to cloudpickling
    TRANSPORTABLE = 1  # Custom TO serialization
    JSONABLE = 2
    TEXT = 3  # Mainly for stdout, stderr, docstrings, etc.
    BYTES = 4  # Just bytes, used for qelectron DB, and other binary data


def serialize_asset(data: Any, data_type: AssetType) -> bytes:
    """
    Serialize the asset data

    Args:
        data: Data to serialize
        data_type: Type of the Asset data to serialize

    Returns:
        Serialized data as bytes

    """

    if data_type == AssetType.OBJECT:
        return cloudpickle.dumps(data)
    elif data_type == AssetType.TRANSPORTABLE:
        return data.serialize()
    elif data_type == AssetType.JSONABLE:
        return json.dumps(data).encode("utf-8")
    elif data_type == AssetType.TEXT:
        return data.encode("utf-8")
    elif data_type == AssetType.BYTES:
        return data
    else:
        raise TypeError(f"Unsupported data type {type(data)}")


def deserialize_asset(data: bytes, data_type: AssetType) -> Any:
    """
    Deserialize the asset data

    Args:
        data: Data to deserialize
        data_type: Type of the Asset data to deserialize

    Returns:
        Deserialized data

    """

    if data_type == AssetType.OBJECT:
        return cloudpickle.loads(data)
    elif data_type == AssetType.TRANSPORTABLE:
        return TransportableObject.deserialize(data)
    elif data_type == AssetType.JSONABLE:
        return json.loads(data.decode("utf-8"))
    elif data_type == AssetType.TEXT:
        return data.decode("utf-8")
    elif data_type == AssetType.BYTES:
        return data
    else:
        raise TypeError("Unsupported data type")


def _sha1_asset(data: bytes) -> str:
    """
    Compute the sha1 checksum of the asset data

    Args:
        data: Data to compute checksum for

    Returns:
        sha1 checksum of the data

    """

    return hashlib.sha1(data).hexdigest()


def save_asset(data: Any, data_type: AssetType, storage_path: str, filename: str) -> AssetSchema:
    """
    Save the asset data to the storage path

    Args:
        data: Data to save
        data_type: Type of the Asset data to save
        storage_path: Path to save the data to
        filename: Name of the file to save the data to

    Returns:
        AssetSchema object containing metadata about the saved data

    """

    scheme = "file"

    if data is None:
        return AssetSchema(size=0)

    serialized = serialize_asset(data, data_type)
    digest = _sha1_asset(serialized)
    path = Path(storage_path) / filename
    path = path.resolve()
    with open(path, "wb") as f:
        f.write(serialized)
    uri = f"{scheme}://{path}"
    return AssetSchema(digest_alg=CHECKSUM_ALGORITHM, digest=digest, size=len(serialized), uri=uri)


def load_asset(asset_meta: AssetSchema, data_type: AssetType) -> Any:
    """
    Load the asset data from the storage path

    Args:
        asset_meta: Metadata about the asset to load
        data_type: Type of the Asset data to load

    Returns:
        Asset data

    """

    scheme_prefix = "file://"
    uri = asset_meta.uri

    if not uri:
        return None

    path = uri[len(scheme_prefix) :] if uri.startswith(scheme_prefix) else uri

    with open(path, "rb") as f:
        data = f.read()
    return deserialize_asset(data, data_type)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for defining a Dask executor that submits the input python function in a dask cluster
and waits for execution to finish then returns the result.

This is a plugin executor module; it is loaded if found and properly structured.
"""

import asyncio
import json
import os
from enum import Enum
from typing import Any, Callable, Dict, List, Literal, Optional

from dask.distributed import CancelledError, Client, Future
from pydantic import BaseModel

from covalent._shared_files import TaskRuntimeError, logger

# Relative imports are not allowed in executor plugins
from covalent._shared_files.config import get_config
from covalent._shared_files.exceptions import TaskCancelledError
from covalent._shared_files.util_classes import RESULT_STATUS, Status
from covalent._shared_files.utils import format_server_url
from covalent.executor.base import AsyncBaseExecutor
from covalent.executor.schemas import ResourceMap, TaskSpec, TaskUpdate
from covalent.executor.utils.wrappers import io_wrapper as dask_wrapper
from covalent.executor.utils.wrappers import run_task_group_alt

# The plugin class name must be given by the executor_plugin_name attribute:
EXECUTOR_PLUGIN_NAME = "DaskExecutor"

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_EXECUTOR_PLUGIN_DEFAULTS = {
    "log_stdout": "stdout.log",
    "log_stderr": "stderr.log",
    "cache_dir": os.path.join(
        os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"), "covalent"
    ),
    "workdir": os.path.join(
        os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
        "covalent",
        "workdir",
    ),
    "create_unique_workdir": False,
}

# See https://github.com/dask/distributed/issues/5667
_clients = {}

# See
# https://stackoverflow.com/questions/62164283/why-do-my-dask-futures-get-stuck-in-pending-and-never-finish
_futures = {}

MANAGED_EXECUTION = os.environ.get("COVALENT_USE_OLD_DASK") != "1"

# Dictionary to map Dask clients to their scheduler addresses
_address_client_map = {}


# Valid terminal statuses
class StatusEnum(str, Enum):
    CANCELLED = str(RESULT_STATUS.CANCELLED)
    COMPLETED = str(RESULT_STATUS.COMPLETED)
    FAILED = str(RESULT_STATUS.FAILED)
    READY = "READY"


class ReceiveModel(BaseModel):
    status: StatusEnum


class DaskExecutor(AsyncBaseExecutor):
    """
    Dask executor class that submits the input function to a running LOCAL dask cluster.
    """

    SUPPORTS_MANAGED_EXECUTION = MANAGED_EXECUTION

    def __init__(
        self,
        scheduler_address: str = "",
        log_stdout: str = "stdout.log",
        log_stderr: str = "stderr.log",
        conda_env: str = "",
        cache_dir: str = "",
        current_env_on_conda_fail: bool = False,
        workdir: str = "",
        create_unique_workdir: Optional[bool] = None,
    ) -> None:
        if not cache_dir:
            cache_dir = _EXECUTOR_PLUGIN_DEFAULTS["cache_dir"]

        if not workdir:
            try:
                workdir = get_config("executors.dask.workdir")
            except KeyError:
                workdir = _EXECUTOR_PLUGIN_DEFAULTS["workdir"]
                debug_msg = f"Couldn't find `executors.dask.workdir` in config, using default value {workdir}."
                app_log.debug(debug_msg)

        if create_unique_workdir is None:
            try:
                create_unique_workdir = get_config("executors.dask.create_unique_workdir")
            except KeyError:
                create_unique_workdir = _EXECUTOR_PLUGIN_DEFAULTS["create_unique_workdir"]
                debug_msg = f"Couldn't find `executors.dask.create_unique_workdir` in config, using default value {create_unique_workdir}."
                app_log.debug(debug_msg)

        super().__init__(
            log_stdout,
            log_stderr,
            cache_dir,
            conda_env,
            current_env_on_conda_fail,
        )

        self.workdir = workdir
        self.create_unique_workdir = create_unique_workdir
        self.scheduler_address = scheduler_address

    async def run(self, function: Callable, args: List, kwargs: Dict, task_metadata: Dict):
        """Submit the function and inputs to the dask cluster"""

        if not self.scheduler_address:
            try:
                self.scheduler_address = get_config("dask.scheduler_address")
            except KeyError:
                app_log.debug(
                    "No dask scheduler address found in config. Address must be set manually."
                )

        if await self.get_cancel_requested():
            app_log.debug("Task has cancelled")
            raise TaskCancelledError

        dispatch_id = task_metadata["dispatch_id"]
        node_id = task_metadata["node_id"]

        dask_client = _address_client_map.get(self.scheduler_address)

        if not dask_client:
            dask_client = Client(address=self.scheduler_address, asynchronous=True)
            _address_client_map[self.scheduler_address] = dask_client
            await dask_client

        if self.create_unique_workdir:
            current_workdir = os.path.join(self.workdir, dispatch_id, f"node_{node_id}")
        else:
            current_workdir = self.workdir

        future = dask_client.submit(dask_wrapper, function, args, kwargs, current_workdir)
        await self.set_job_handle(future.key)
        app_log.debug(f"Submitted task {node_id} to dask with key {future.key}")

        try:
            result, worker_stdout, worker_stderr, tb = await future
        except CancelledError as e:
            raise TaskCancelledError() from e

        print(worker_stdout, end="", file=self.task_stdout)
        print(worker_stderr, end="", file=self.task_stderr)

        if tb:
            print(tb, end="", file=self.task_stderr)
            raise TaskRuntimeError(tb)

        # FIX: need to get stdout and stderr from dask worker and print them
        return result

    async def cancel(self, task_metadata: Dict, job_handle) -> Literal[True]:
        """
        Cancel the task being executed by the dask executor currently

        Arg(s)
            task_metadata: Metadata associated with the task
            job_handle: Key assigned to the job by Dask

        Return(s)
            True by default
        """

        if not self.scheduler_address:
            try:
                self.scheduler_address = get_config("dask.scheduler_address")
            except KeyError:
                app_log.debug(
                    "No dask scheduler address found in config. Address must be set manually."
                )

        dask_client = _address_client_map.get(self.scheduler_address)

        if not dask_client:
            dask_client = Client(address=self.scheduler_address, asynchronous=True)
            await asyncio.wait_for(dask_client, timeout=5)

        fut: Future = Future(key=job_handle, client=dask_client)

        # https://stackoverflow.com/questions/46278692/dask-distributed-how-to-cancel-tasks-submitted-with-fire-and-forget
        await dask_client.cancel([fut], asynchronous=True, force=True)
        app_log.debug(f"Cancelled future with key {job_handle}")
        return True

    async def send(
        self,
        task_specs: List[TaskSpec],
        resources: ResourceMap,
        task_group_metadata: dict,
    ):
        # Assets are assumed to be accessible by the compute backend
        # at the provided URIs

        # The Asset Manager is responsible for uploading all assets
        # Returns a job handle (should be JSONable)

        if not self.scheduler_address:
            try:
                self.scheduler_address = get_config("dask.scheduler_address")
            except KeyError:
                app_log.debug(
                    "No dask scheduler address found in config. Address must be set manually."
                )

        dask_client = Client(address=self.scheduler_address, asynchronous=True)
        await asyncio.wait_for(dask_client, timeout=5)

        dispatch_id = task_group_metadata["dispatch_id"]
        task_ids = task_group_metadata["node_ids"]
        gid = task_group_metadata["task_group_id"]
        output_uris = []
        for node_id in task_ids:
            result_uri = os.path.join(self.cache_dir, f"result_{dispatch_id}-{node_id}.pkl")
            stdout_uri = os.path.join(self.cache_dir, f"stdout_{dispatch_id}-{node_id}.txt")
            stderr_uri = os.path.join(self.cache_dir, f"stderr_{dispatch_id}-{node_id}.txt")
            qelectron_db_uri = os.path.join(
                self.cache_dir, f"qelectron_db_{dispatch_id}-{node_id}.mdb"
            )
            output_uris.append((result_uri, stdout_uri, stderr_uri, qelectron_db_uri))

        server_url = format_server_url()

        key = f"dask_job_{dispatch_id}:{gid}"

        await self.set_job_handle(key)

        future = dask_client.submit(
            run_task_group_alt,
            list(map(lambda t: t.model_dump(), task_specs)),
            resources.model_dump(),
            output_uris,
            self.cache_dir,
            task_group_metadata,
            server_url,
            key=key,
        )

        _clients[key] = dask_client
        _futures[key] = future

        return future.key

    async def poll(self, task_group_metadata: Dict, poll_data: Any):
        fut = _futures.pop(poll_data)
        app_log.debug(f"Future {fut}")
        try:
            await fut
        except CancelledError as e:
            raise TaskCancelledError() from e

        _clients.pop(poll_data)

        return {"status": StatusEnum.READY.value}

    async def receive(self, task_group_metadata: Dict, data: Any) -> List[TaskUpdate]:
        # Job should have reached a terminal state by the time this is invoked.
        dispatch_id = task_group_metadata["dispatch_id"]
        task_ids = task_group_metadata["node_ids"]

        task_results = []

        if not data:
            terminal_status = RESULT_STATUS.CANCELLED
        else:
            received = ReceiveModel.model_validate(data)
            terminal_status = Status(received.status.value)

        for task_id in task_ids:
            # TODO: Handle the case where the job was cancelled before the task started running
            app_log.debug(f"Receive called for task {dispatch_id}:{task_id} with data {data}")

            if terminal_status == RESULT_STATUS.CANCELLED:
                output_uri = ""
                output_size = 0
                stdout_uri = ""
                stdout_size = 0
                stderr_uri = ""
                stderr_size = 0
                qelectron_db_uri = ""
                qelectron_db_size = 0

            else:
                result_path = os.path.join(self.cache_dir, f"result-{dispatch_id}:{task_id}.json")
                with open(result_path, "r") as f:
                    result_summary = json.load(f)
                    node_id = result_summary["node_id"]
                    output_uri = result_summary["output"]["uri"]
                    output_size = result_summary["output"]["size"]
                    stdout_uri = result_summary["stdout"]["uri"]
                    stdout_size = result_summary["stdout"]["size"]
                    stderr_uri = result_summary["stderr"]["uri"]
                    stderr_size = result_summary["stderr"]["size"]
                    qelectron_db_uri = result_summary["qelectron_db"]["uri"]
                    qelectron_db_size = result_summary["qelectron_db"]["size"]
                    exception_raised = result_summary["exception_occurred"]

                terminal_status = (
                    RESULT_STATUS.FAILED if exception_raised else RESULT_STATUS.COMPLETED
                )

            task_result = {
                "dispatch_id": dispatch_id,
                "node_id": task_id,
                "status": terminal_status,
                "assets": {
                    "output": {
                        "remote_uri": output_uri,
                        "size": output_size,
                    },
                    "stdout": {
                        "remote_uri": stdout_uri,
                        "size": stdout_size,
                    },
                    "stderr": {
                        "remote_uri": stderr_uri,
                        "size": stderr_size,
                    },
                    "qelectron_db": {
                        "remote_uri": qelectron_db_uri,
                        "size": qelectron_db_size,
                    },
                },
            }

            task_results.append(TaskUpdate(**task_result))

        app_log.debug(f"Returning results for tasks {dispatch_id}:{task_ids}")
        return task_results

    def get_upload_uri(self, task_group_metadata: Dict, object_key: str):
        dispatch_id = task_group_metadata["dispatch_id"]
        task_group_id = task_group_metadata["task_group_id"]

        filename = f"asset_{dispatch_id}-{task_group_id}_{object_key}.pkl"
        return os.path.join("file://", self.cache_dir, filename)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper classes for the dispatcher
"""

from .store import _DictStore, _KeyValueBase


def _pending_parents_key(dispatch_id: str, node_id: int):
    return f"pending-parents-{dispatch_id}:{node_id}"


def _unresolved_tasks_key(dispatch_id: str):
    return f"unresolved-{dispatch_id}"


def _task_groups_key(dispatch_id: str, task_group_id: int):
    return f"task-groups-{dispatch_id}:{task_group_id}"


class _UnresolvedTasksCache:
    def __init__(self, store: _KeyValueBase = _DictStore()):
        self._store = store

    async def get_unresolved(self, dispatch_id: str):
        key = _unresolved_tasks_key(dispatch_id)
        return await self._store.get(key)

    async def set_unresolved(self, dispatch_id: str, val: int):
        key = _unresolved_tasks_key(dispatch_id)
        await self._store.insert(key, val)

    async def increment(self, dispatch_id: str, interval: int = 1):
        key = _unresolved_tasks_key(dispatch_id)
        return await self._store.increment(key, interval)

    async def decrement(self, dispatch_id: str):
        key = _unresolved_tasks_key(dispatch_id)
        return await self._store.increment(key, -1)

    async def remove(self, dispatch_id: str):
        key = _unresolved_tasks_key(dispatch_id)
        await self._store.remove(key)


class _PendingParentsCache:
    def __init__(self, store: _KeyValueBase = _DictStore()):
        self._store = store

    async def get_pending(self, dispatch_id: str, task_group_id: int):
        key = _pending_parents_key(dispatch_id, task_group_id)
        return await self._store.get(key)

    async def set_pending(self, dispatch_id: str, task_group_id: int, val: int):
        key = _pending_parents_key(dispatch_id, task_group_id)
        await self._store.insert(key, val)

    async def decrement(self, dispatch_id: str, task_group_id: int):
        key = _pending_parents_key(dispatch_id, task_group_id)
        return await self._store.increment(key, -1)

    async def remove(self, dispatch_id: str, task_group_id: int):
        key = _pending_parents_key(dispatch_id, task_group_id)
        await self._store.remove(key)


class _SortedTaskGroups:
    def __init__(self, store: _KeyValueBase = _DictStore()):
        self._store = store

    async def get_task_group(self, dispatch_id: str, task_group_id: int):
        key = _task_groups_key(dispatch_id, task_group_id)
        return await self._store.get(key)

    async def set_task_group(self, dispatch_id: str, task_group_id: int, sorted_nodes: list):
        key = _task_groups_key(dispatch_id, task_group_id)
        await self._store.insert(key, sorted_nodes)

    async def remove(self, dispatch_id: str, task_group_id: int):
        key = _task_groups_key(dispatch_id, task_group_id)
        await self._store.remove(key)


_pending_parents = _PendingParentsCache()
_unresolved_tasks = _UnresolvedTasksCache()
_sorted_task_groups = _SortedTaskGroups()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DB-backed edge"""


from typing import Dict

from .._db import models
from .controller import Record
from .db_interfaces.edge_utils import _to_edge_attrs, _to_endpoints


class Edge:
    def __init__(self, record: models.ElectronDependency, uid_node_id_map: Dict):
        self.source, self.target = _to_endpoints(record, uid_node_id_map)
        self.attrs = _to_edge_attrs(record)


class ElectronDependency(Record[models.ElectronDependency]):
    model = models.ElectronDependency



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions providing programmatic access to Covalent CLI commands."""
import subprocess
from typing import List, Optional

import psutil

from .._shared_files import logger
from .._shared_files.config import get_config

__all__ = ["is_covalent_running", "covalent_start", "covalent_stop"]


app_log = logger.app_log

_MISSING_SERVER_WARNING = "Covalent has not been installed with the server component."


def _call_cli_command(cmd: List[str], *, quiet: bool = False) -> subprocess.CompletedProcess:
    """
    Call a CLI command with the specified kwargs.

    Args:
        func: The CLI command to call.
        quiet: Suppress stdout. Defaults to :code:`False`.
    """

    if quiet:
        return subprocess.run(
            cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )

    return subprocess.run(cmd, check=True)


def is_covalent_running() -> bool:
    """
    Check if the Covalent server is running.

    Returns:
        :code:`True` if the Covalent server is in a ready state, :code:`False` otherwise.
    """
    try:
        from covalent_dispatcher._cli.service import _read_pid

        pid = _read_pid(get_config("dispatcher.cache_dir") + "/ui.pid")
        return (
            pid != -1
            and psutil.pid_exists(pid)
            and get_config("dispatcher.address") != ""
            and get_config("dispatcher.port") != ""
        )

    except ModuleNotFoundError:
        # If the covalent_dispatcher is not installed, assume Covalent is not running.
        app_log.warning(_MISSING_SERVER_WARNING)
        return False


def covalent_start(
    develop: bool = False,
    port: Optional[str] = None,
    mem_per_worker: Optional[str] = None,
    workers: Optional[int] = None,
    threads_per_worker: Optional[int] = None,
    ignore_migrations: bool = False,
    no_cluster: bool = False,
    no_triggers: bool = False,
    triggers_only: bool = False,
    *,
    quiet: bool = False,
) -> None:
    """
    Start the Covalent server. Wrapper for the :code:`covalent start` CLI command.
    This function returns immediately if the local Covalent server is already running.

    Args:
        develop: Start local server in develop mode. Defaults to :code:`False`.
        port: Local server port number. Defaults to :code:`"48008"`.
        mem_per_worker: Memory limit per worker in GB. Defaults to auto.
        workers: Number of Dask workers. Defaults to 8.
        threads_per_worker: Number of threads per Dask worker. Defaults to 1.
        ignore_migrations: Start server without database migrations. Defaults to :code:`False`.
        no_cluster: Start server without Dask cluster. Defaults to :code:`False`.
        no_triggers: Start server without a triggers server. Defaults to :code:`False`.
        triggers_only: Start only the triggers server. Defaults to :code:`False`.
        quiet: Suppress stdout. Defaults to :code:`False`.
    """

    if is_covalent_running():
        msg = "Covalent server is already running."
        if not quiet:
            print(msg)

        app_log.debug(msg)
        return

    flags = {
        "--develop": develop,
        "--ignore-migrations": ignore_migrations,
        "--no-cluster": no_cluster,
        "--no-triggers": no_triggers,
        "--triggers-only": triggers_only,
    }

    args = {
        "--port": port or get_config("dispatcher.port"),
        "--mem-per-worker": mem_per_worker or get_config("dask.mem_per_worker"),
        "--workers": workers or get_config("dask.num_workers"),
        "--threads-per-worker": threads_per_worker or get_config("dask.threads_per_worker"),
    }

    cmd = ["covalent", "start"]
    cmd.extend(flag for flag, value in flags.items() if value)

    for arg, value in args.items():
        cmd.extend((arg, str(value)))

    # Run the `covalent start [OPTIONS]` command.
    app_log.debug("Starting Covalent server programmatically...")
    _call_cli_command(cmd, quiet=quiet)


def covalent_stop(*, quiet: bool = False) -> None:
    """
    Stop the Covalent server. Wrapper for the :code:`covalent stop` CLI command.
    This function returns immediately if the local Covalent server is not running.

    Args:
        quiet: Suppress stdout. Defaults to :code:`False`.
    """

    if not is_covalent_running():
        msg = "Covalent server is not running."
        if not quiet:
            print(msg)

        app_log.debug(msg)
        return

    # Run the `covalent stop` command.
    app_log.debug("Stopping Covalent server programmatically...")
    _call_cli_command(["covalent", "stop"], quiet=quiet)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass
from typing import NamedTuple


@dataclass
class Status:
    STATUS: str

    def __bool__(self):
        """
        Return True if the status is not "NEW_OBJECT"
        """

        return self.STATUS != "NEW_OBJECT"

    def __str__(self) -> str:
        return self.STATUS

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, self.__class__):
            return self.STATUS == __value.STATUS
        elif isinstance(__value, str):
            return self.STATUS == __value
        return False

    def __ne__(self, __value: object) -> bool:
        return not self.__eq__(__value)


class RESULT_STATUS:
    NEW_OBJECT = Status("NEW_OBJECT")
    STARTING = Status("STARTING")  # Dispatch level
    PENDING_REUSE = Status("PENDING_REUSE")  # For redispatch in the new dispatcher design
    PENDING_REPLACEMENT = Status(
        "PENDING_REPLACEMENT"
    )  # For redispatch in the new dispatcher design
    COMPLETED = Status("COMPLETED")
    POSTPROCESSING = Status("POSTPROCESSING")
    PENDING_POSTPROCESSING = Status("PENDING_POSTPROCESSING")
    POSTPROCESSING_FAILED = Status("POSTPROCESSING_FAILED")
    FAILED = Status("FAILED")
    RUNNING = Status("RUNNING")
    CANCELLED = Status("CANCELLED")
    DISPATCHING = Status("DISPATCHING")
    DISPATCHING_SUBLATTICE = Status("DISPATCHING")

    @staticmethod
    def is_terminal(status):
        return str(status) in TERMINAL_STATUSES


TERMINAL_STATUSES = {
    str(RESULT_STATUS.COMPLETED),
    str(RESULT_STATUS.FAILED),
    str(RESULT_STATUS.CANCELLED),
}


class DispatchInfo(NamedTuple):
    """
    Information about a dispatch to be shared to a task post dispatch.

    Attributes:
        dispatch_id: Dispatch id of the dispatch.
    """

    dispatch_id: str



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Define the custom Pennylane device that interacts with Covalent's Quantum Executors.
"""

from typing import Sequence

from pennylane import QubitDevice
from pennylane import numpy as np
from pennylane.devices.default_qubit import DefaultQubit

from ..quantum.qclient.core import middleware


class QEDevice(QubitDevice):

    """
    The purpose of this device is to redirect circuit execution through Covalent's
    Quantum Executors and to enable asynchronous execution.

    All QElectrons (which are QNodes) implicitly use this device.
    """

    name = "QEDevice"
    short_name = "qelectron_device"
    pennylane_requires = ">=0.29.1"
    version = "0.0.1"
    author = "aq"

    operations = DefaultQubit.operations
    observables = DefaultQubit.observables

    def __init__(
        self,
        wires=1,
        shots=None,
        *,
        r_dtype=np.float64,  # pylint: disable=no-member
        c_dtype=np.complex128,  # pylint: disable=no-member
        analytic=None,
        executors=None,
        qelectron_info=None,
    ):
        super().__init__(wires, shots, r_dtype=r_dtype, c_dtype=c_dtype, analytic=analytic)

        self._async_run = False
        self._batch_id = None
        self.executors = executors
        self.qelectron_info = qelectron_info

        # This will be set when the QNodeQE is called with args and kwargs.
        self.qnode_specs = None

    def batch_execute(self, circuits):
        """
        Submits circuits to QElectron middleware for execution on chosen backend(s).
        Retrieves and returns results, or returns a dummy result if running asynchronously.
        """

        # Async submit all circuits to middleware.
        batch_id = middleware.run_circuits_async(
            circuits, self.executors, self.qelectron_info, self.qnode_specs
        )

        # Relevant when `run_later` is used to run the circuit.
        # We will retrieve the result later using the future object and the batch_id.
        if self._async_run:
            self._batch_id = batch_id

            # Return a (recognizable) dummy result
            res = [self._asarray([-123.456] * c.output_dim) for c in circuits]
            if isinstance(self.qelectron_info.device_shots, Sequence):
                # Replicate the list for shot vector case.
                return [[res]] * len(self.qelectron_info.device_shots)
            return res

        # Otherwise, get the results from the middleware
        results = middleware.get_results(batch_id)

        return results

    def apply(self, *args, **kwargs):
        # Dummy implementation of abstractmethod on `QubitDevice`
        pass

    @classmethod
    def capabilities(cls):
        capabilities = super().capabilities().copy()
        capabilities.update(supports_broadcasting=True)
        return capabilities



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Request and response models for Electrons"""

from datetime import datetime
from enum import Enum
from typing import Dict, List, Union

from pydantic import BaseModel

from covalent_ui.api.v1.utils.status import Status


class Job(BaseModel):
    job_id: Union[str, None] = None
    start_time: Union[datetime, None] = None
    executor: Union[str, None] = None
    status: Union[str, None] = None


class JobsResponse(BaseModel):
    data: Union[List[Job], None] = None
    msg: Union[str, None] = None


class JobDetails(BaseModel):
    overview: Union[dict, None] = None
    circuit: Union[dict, None] = None
    executor: Union[dict, None] = None


class JobDetailsResponse(BaseModel):
    data: Union[JobDetails, None, Dict] = None
    msg: Union[str, None] = None


class ElectronResponse(BaseModel):
    """Electron Response Model"""

    id: Union[int, None] = None
    node_id: Union[int, None] = None
    parent_lattice_id: Union[int, None] = None
    type: Union[str, None] = None
    storage_path: Union[str, None] = None
    name: Union[str, None] = None
    status: Union[str, None] = None
    started_at: Union[datetime, None] = None
    ended_at: Union[datetime, None] = None
    runtime: Union[int, float, None] = None
    description: Union[str, None] = None
    qelectron_data_exists: bool = False
    qelectron: Union[dict, None] = None


class ElectronFileResponse(BaseModel):
    """Electron Response Model"""

    data: Union[str, None] = None
    python_object: Union[str, None] = None


class ElectronExecutorResponse(BaseModel):
    """Electron File Response Model"""

    executor_name: Union[str, None] = None
    executor_details: Union[dict, None] = None


class ElectronErrorResponse(BaseModel):
    """Electron Error Response Model"""

    data: Union[str, None] = None


class ElectronFunctionResponse(BaseModel):
    """Electron Function Response Model"""

    data: Union[str, None] = None


class ElectronResponseModel(BaseModel):
    """Electron Response Validation"""

    id: int
    parent_lattice_id: int
    transport_graph_node_id: int
    type: str
    name: str
    status: Status
    executor: str
    created_at: str
    started_at: str
    completed_at: str
    updated_at: str


class ElectronFileOutput(str, Enum):
    """Electron file output"""

    FUNCTION_STRING = "function_string"
    FUNCTION = "function"
    EXECUTOR = "executor"
    RESULT = "result"
    VALUE = "value"
    STDOUT = "stdout"
    HOOKS = "hooks"
    ERROR = "error"
    INFO = "info"
    INPUTS = "inputs"



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add jobs db table

Revision ID: f64ecaa040d5
Revises: 97202f5f47cb
Create Date: 2023-02-07 14:52:43.475708

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
# pragma: allowlist nextline secret
revision = "f64ecaa040d5"
# pragma: allowlist nextline secret
down_revision = "97202f5f47cb"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "jobs",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("cancel_requested", sa.Boolean(), nullable=False),
        sa.Column("cancel_successful", sa.Boolean(), nullable=False),
        sa.Column("job_handle", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("electron_dependency", schema=None) as batch_op:
        batch_op.create_foreign_key("electron_link", "electrons", ["parent_electron_id"], ["id"])

    with op.batch_alter_table("electrons", schema=None) as batch_op:
        batch_op.add_column(sa.Column("job_id", sa.Integer(), nullable=False))
        batch_op.create_foreign_key("job_id_link", "jobs", ["job_id"], ["id"])

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("electrons", schema=None) as batch_op:
        batch_op.drop_constraint("job_id_link", type_="foreignkey")
        batch_op.drop_column("job_id")

    with op.batch_alter_table("electron_dependency", schema=None) as batch_op:
        batch_op.drop_constraint("electron_link", type_="foreignkey")

    op.drop_table("jobs")
    # ### end Alembic commands ###



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QNode and QElectron information containers."""

from typing import Any, Dict, Optional, Sequence, Union

from pydantic import BaseModel  # pylint: disable=no-name-in-module


class QNodeSpecs(BaseModel):
    """
    A container for the specifications of a QNode generated by `qml.specs`.
    """

    gate_sizes: Dict[int, int]
    gate_types: Dict[str, int]
    num_operations: int
    num_observables: int
    num_diagonalizing_gates: int
    num_used_wires: int
    depth: int
    num_trainable_params: int = None
    num_device_wires: int
    device_name: str
    diff_method: Optional[str] = None
    expansion_strategy: str
    gradient_options: Dict[str, int]
    interface: Optional[str] = None
    gradient_fn: Any = None  # can be string or `qml.gradients.gradient_transform`
    num_gradient_executions: Any = 0
    num_parameter_shift_executions: int = None


class QElectronInfo(BaseModel):
    """
    A container for global and QNode device settings.
    """

    name: str
    description: Optional[str] = None
    device_name: str  # name of the original device, e.g. "default.qubit"
    device_import_path: str  # used to inherit type converters and other methods
    device_shots: Union[
        None, int, Sequence[int], Sequence[Union[int, Sequence[int]]]
    ] = None  # optional default for execution devices
    device_shots_type: Any = None
    device_wires: int  # this can not be reliably inferred from tapes alone
    pennylane_active_return: bool  # client-side status of `pennylane.active_return()`



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Functions to transform Lattice -> LatticeSchema"""


from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.schemas.asset import AssetSchema
from covalent._shared_files.schemas.result import (
    ASSET_KEYS,
    METADATA_KEYS,
    ResultAssets,
    ResultMetadata,
    ResultSchema,
)
from covalent._shared_files.utils import format_server_url
from covalent_dispatcher._dal.electron import Electron
from covalent_dispatcher._dal.result import Result, get_result_object

from ..utils.uri_filters import AssetScope, URIFilterPolicy, filter_asset_uri
from .lattice import export_lattice

METADATA_KEYS_TO_OMIT = {"num_nodes"}
SERVER_URL = format_server_url(get_config("dispatcher.address"), get_config("dispatcher.port"))
URI_FILTER_POLICY = URIFilterPolicy[get_config("dispatcher.data_uri_filter_policy")]

app_log = logger.app_log


# res is assumed to represent a full db record
def _export_result_meta(res: Result) -> ResultMetadata:
    metadata_kwargs = {
        key: res.get_metadata(key, None, refresh=False)
        for key in METADATA_KEYS
        if key not in METADATA_KEYS_TO_OMIT
    }
    return ResultMetadata(**metadata_kwargs)


def _populate_assets(res: Result):
    """Prepopulate the asset maps"""

    # Compute mapping from electron_id -> transport_graph_node_id

    tg = res.lattice.transport_graph
    g = tg.get_internal_graph_copy()
    all_nodes = tg.get_nodes(node_ids=list(g.nodes))

    eid_node_id_map = {node._electron_id: node.node_id for node in all_nodes}

    with res.session() as session:
        # Workflow scope
        workflow_assets = Result.get_linked_assets(
            session,
            fields=[],
            equality_filters={"id": res.metadata.primary_key},
            membership_filters={},
        )
        # Electron scope

        node_assets = Electron.get_linked_assets(
            session,
            fields=[],
            equality_filters={"parent_lattice_id": res.metadata.primary_key},
            membership_filters={},
        )

    for rec in workflow_assets:
        res.assets[rec["key"]] = rec["asset"]

    for key, val in res.assets.items():
        res.lattice.assets[key] = val

    for rec in node_assets:
        node = tg.get_node(eid_node_id_map[rec["meta_id"]])
        node.assets[rec["key"]] = rec["asset"]


def _export_result_assets(res: Result) -> ResultAssets:
    manifests = {}
    for asset_key in ASSET_KEYS:
        asset = res.assets[asset_key]
        size = asset.size
        digest_alg = asset.digest_alg
        digest = asset.digest
        scheme = asset.storage_type.value
        remote_uri = f"{scheme}://{asset.storage_path}/{asset.object_key}"
        manifests[asset_key] = AssetSchema(
            remote_uri=remote_uri, size=size, digest_alg=digest_alg, digest=digest
        )

    return ResultAssets(**manifests)


def export_result(res: Result) -> ResultSchema:
    """Export a Result object"""
    dispatch_id = res.dispatch_id
    metadata = _export_result_meta(res)

    _populate_assets(res)

    assets = _export_result_assets(res)
    lattice = export_lattice(res.lattice)

    # Filter asset URIs

    return _filter_remote_uris(ResultSchema(metadata=metadata, assets=assets, lattice=lattice))


def _filter_remote_uris(manifest: ResultSchema) -> ResultSchema:
    dispatch_id = manifest.metadata.dispatch_id

    # Workflow-level
    for key, asset in manifest.assets:
        filtered_uri = filter_asset_uri(
            URI_FILTER_POLICY, asset.remote_uri, {}, AssetScope.DISPATCH, dispatch_id, None, key
        )
        asset.remote_uri = filtered_uri

    for key, asset in manifest.lattice.assets:
        filtered_uri = filter_asset_uri(
            URI_FILTER_POLICY, asset.remote_uri, {}, AssetScope.LATTICE, dispatch_id, None, key
        )
        asset.remote_uri = filtered_uri

    # Now filter each node
    tg = manifest.lattice.transport_graph
    for node in tg.nodes:
        for key, asset in node.assets:
            filtered_uri = filter_asset_uri(
                URI_FILTER_POLICY, asset.remote_uri, {}, AssetScope.NODE, dispatch_id, node.id, key
            )
            asset.remote_uri = filtered_uri

    return manifest


def export_result_manifest(dispatch_id: str) -> ResultSchema:
    srv_res = get_result_object(dispatch_id, bare=False)
    return export_result(srv_res)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import contextlib
import os
from pathlib import Path
from typing import Dict, List, Optional

from furl import furl
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from .._api.apiclient import CovalentAPIClient
from .._serialize.common import load_asset
from .._serialize.electron import ASSET_FILENAME_MAP as ELECTRON_ASSET_FILENAMES
from .._serialize.electron import ASSET_TYPES as ELECTRON_ASSET_TYPES
from .._serialize.lattice import ASSET_FILENAME_MAP as LATTICE_ASSET_FILENAMES
from .._serialize.lattice import ASSET_TYPES as LATTICE_ASSET_TYPES
from .._serialize.result import ASSET_FILENAME_MAP as RESULT_ASSET_FILENAMES
from .._serialize.result import ASSET_TYPES as RESULT_ASSET_TYPES
from .._serialize.result import deserialize_result
from .._shared_files import logger
from .._shared_files.config import get_config
from .._shared_files.exceptions import MissingLatticeRecordError
from .._shared_files.schemas.asset import AssetSchema
from .._shared_files.schemas.result import ResultSchema
from .._shared_files.utils import copy_file_locally, format_server_url
from .result import Result
from .wait import EXTREME

app_log = logger.app_log
log_stack_info = logger.log_stack_info


SDK_NODE_META_KEYS = {
    "executor",
    "executor_data",
    "hooks",
}

SDK_LAT_META_KEYS = {
    "executor",
    "executor_data",
    "workflow_executor",
    "workflow_executor_data",
    "hooks",
}

DEFERRED_KEYS = {
    "output",
    "result",
    "qelectron_db",
}


def _delete_result(
    dispatch_id: str,
    results_dir: str = None,
    remove_parent_directory: bool = False,
) -> None:
    """
    Internal function to delete the result.

    Args:
        dispatch_id: The dispatch id of the result.
        results_dir: The directory where the results are stored in dispatch id named folders.
        remove_parent_directory: Status of whether to delete the parent directory when removing the result.

    Returns:
        None

    Raises:
        FileNotFoundError: If the result file is not found.
    """

    if results_dir is None:
        results_dir = os.environ.get("COVALENT_DATA_DIR") or get_config("dispatcher.results_dir")

    import shutil

    result_folder_path = os.path.join(results_dir, f"{dispatch_id}")

    if os.path.exists(result_folder_path):
        shutil.rmtree(result_folder_path, ignore_errors=True)

    with contextlib.suppress(OSError):
        os.rmdir(results_dir)

    if remove_parent_directory:
        shutil.rmtree(results_dir, ignore_errors=True)


def cancel(dispatch_id: str, task_ids: List[int] = None, dispatcher_addr: str = None) -> str:
    """
    Cancel a running dispatch.

    Args:
        dispatch_id: The dispatch id of the dispatch to be cancelled.
        task_ids: Optional, list of task ids to cancel within the workflow
        dispatcher_addr: Dispatcher server address, if None then defaults to the address set in Covalent's config.

    Returns:
        Cancellation response
    """

    if dispatcher_addr is None:
        dispatcher_addr = format_server_url()

    if task_ids is None:
        task_ids = []

    api_client = CovalentAPIClient(dispatcher_addr)
    endpoint = f"/api/v2/dispatches/{dispatch_id}/status"

    if isinstance(task_ids, int):
        task_ids = [task_ids]

    body = {"status": "CANCELLED", "task_ids": task_ids}
    r = api_client.put(endpoint, json=body)
    return r.content.decode("utf-8").strip().replace('"', "")


# Multi-part


def _get_result_export_from_dispatcher(
    dispatch_id: str,
    wait: bool = False,
    status_only: bool = False,
    dispatcher_addr: str = None,
) -> Dict:
    """
    Internal function to get the results of a dispatch from the server without checking if it is ready to read.

    Args:
        dispatch_id: The dispatch id of the result.
        wait: Controls how long the method waits for the server to return a result. If False, the method will not wait and will return the current status of the workflow. If True, the method will wait for the result to finish and keep retrying for sys.maxsize.
        status_only: If true, only returns result status, not the full result object, default is False.
        dispatcher_addr: Dispatcher server address, defaults to the address set in covalent.config.

    Returns:
        The result object from the server.

    Raises:
        MissingLatticeRecordError: If the result is not found.
    """

    if dispatcher_addr is None:
        dispatcher_addr = format_server_url()

    retries = int(EXTREME) if wait else 5

    adapter = HTTPAdapter(max_retries=Retry(total=retries, backoff_factor=1))
    api_client = CovalentAPIClient(dispatcher_addr, adapter=adapter, auto_raise=False)

    endpoint = f"/api/v2/dispatches/{dispatch_id}"
    response = api_client.get(
        endpoint,
        params={"wait": wait, "status_only": status_only},
    )
    if response.status_code == 404:
        raise MissingLatticeRecordError
    response.raise_for_status()
    export = response.json()
    return export


# Function to download default assets
def _get_default_assets(rm: ResultManager):
    for key in RESULT_ASSET_TYPES.keys():
        if key not in DEFERRED_KEYS:
            rm.download_result_asset(key)
            rm.load_result_asset(key)

    for key in LATTICE_ASSET_TYPES.keys():
        if key not in DEFERRED_KEYS:
            rm.download_lattice_asset(key)
            rm.load_lattice_asset(key)

    tg = rm.result_object.lattice.transport_graph

    tg.lattice_metadata = rm.result_object.lattice.metadata
    rm.result_object.lattice.__doc__ = rm.result_object.lattice.__dict__.pop("doc")

    for key in ELECTRON_ASSET_TYPES.keys():
        if key not in DEFERRED_KEYS:
            for node_id in tg._graph.nodes:
                rm.download_node_asset(node_id, key)
                rm.load_node_asset(node_id, key)


# Functions for computing local URIs
def get_node_asset_path(results_dir: str, node_id: int, key: str):
    filename = ELECTRON_ASSET_FILENAMES[key]
    return f"{results_dir}/node_{node_id}/{filename}"


def get_lattice_asset_path(results_dir: str, key: str):
    filename = LATTICE_ASSET_FILENAMES[key]
    return f"{results_dir}/{filename}"


def get_result_asset_path(results_dir: str, key: str):
    filename = RESULT_ASSET_FILENAMES[key]
    return f"{results_dir}/{filename}"


# Asset transfers


def download_asset(remote_uri: str, local_path: str, chunk_size: int = 1024 * 1024):
    local_scheme = "file"
    if remote_uri.startswith(local_scheme):
        copy_file_locally(remote_uri, f"file://{local_path}")
    else:
        f = furl(remote_uri)
        scheme = f.scheme
        host = f.host
        port = f.port
        dispatcher_addr = f"{scheme}://{host}:{port}"
        endpoint = str(f.path)
        api_client = CovalentAPIClient(dispatcher_addr)
        r = api_client.get(endpoint, stream=True)
        with open(local_path, "wb") as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                f.write(chunk)


def _download_result_asset(manifest: dict, results_dir: str, key: str):
    remote_uri = manifest["assets"][key]["remote_uri"]
    size = manifest["assets"][key]["size"]

    if size > 0:
        local_path = get_result_asset_path(results_dir, key)
        download_asset(remote_uri, local_path)
        manifest["assets"][key]["uri"] = f"file://{local_path}"


def _download_lattice_asset(manifest: dict, results_dir: str, key: str):
    lattice_assets = manifest["lattice"]["assets"]
    remote_uri = lattice_assets[key]["remote_uri"]
    size = lattice_assets[key]["size"]

    if size > 0:
        local_path = get_lattice_asset_path(results_dir, key)
        download_asset(remote_uri, local_path)
        lattice_assets[key]["uri"] = f"file://{local_path}"


def _download_node_asset(manifest: dict, results_dir: str, node_id: int, key: str):
    node = manifest["lattice"]["transport_graph"]["nodes"][node_id]
    node_assets = node["assets"]
    remote_uri = node_assets[key]["remote_uri"]
    size = node_assets[key]["size"]

    if size > 0:
        local_path = get_node_asset_path(results_dir, node_id, key)
        download_asset(remote_uri, local_path)
        node_assets[key]["uri"] = f"file://{local_path}"


def _load_result_asset(manifest: dict, key: str):
    asset_meta = AssetSchema(**manifest["assets"][key])
    return load_asset(asset_meta, RESULT_ASSET_TYPES[key])


def _load_lattice_asset(manifest: dict, key: str):
    asset_meta = AssetSchema(**manifest["lattice"]["assets"][key])
    return load_asset(asset_meta, LATTICE_ASSET_TYPES[key])


def _load_node_asset(manifest: dict, node_id: int, key: str):
    node = manifest["lattice"]["transport_graph"]["nodes"][node_id]
    asset_meta = AssetSchema(**node["assets"][key])
    return load_asset(asset_meta, ELECTRON_ASSET_TYPES[key])


class ResultManager:
    def __init__(self, manifest: ResultSchema, results_dir: str):
        self.result_object = deserialize_result(manifest)
        self._manifest = manifest.model_dump()
        self._results_dir = results_dir

    def save(self, path: Optional[str] = None):
        if not path:
            path = os.path.join(self._results_dir, "manifest.json")
        with open(path, "w") as f:
            f.write(ResultSchema.model_validate(self._manifest).model_dump_json())

    @staticmethod
    def load(path: str, results_dir: str) -> "ResultManager":
        with open(path, "r") as f:
            manifest_json = f.read()

        return ResultManager(ResultSchema.parse_raw(manifest_json), results_dir)

    def download_result_asset(self, key: str):
        _download_result_asset(self._manifest, self._results_dir, key)

    def download_lattice_asset(self, key: str):
        _download_lattice_asset(self._manifest, self._results_dir, key)

    def download_node_asset(self, node_id: int, key: str):
        _download_node_asset(self._manifest, self._results_dir, node_id, key)

    def load_result_asset(self, key: str):
        data = _load_result_asset(self._manifest, key)
        self.result_object.__dict__[f"_{key}"] = data

    def load_lattice_asset(self, key: str):
        data = _load_lattice_asset(self._manifest, key)
        if key in SDK_LAT_META_KEYS:
            self.result_object.lattice.metadata[key] = data
        else:
            self.result_object.lattice.__dict__[key] = data

    def load_node_asset(self, node_id: int, key: str):
        data = _load_node_asset(self._manifest, node_id, key)
        tg = self.result_object.lattice.transport_graph
        if key in SDK_NODE_META_KEYS:
            node_meta = tg.get_node_value(node_id, "metadata")
            node_meta[key] = data
        else:
            tg.set_node_value(node_id, key, data)

    @staticmethod
    def from_dispatch_id(
        dispatch_id: str,
        results_dir: str,
        wait: bool = False,
        dispatcher_addr: str = None,
    ) -> "ResultManager":
        export = _get_result_export_from_dispatcher(
            dispatch_id, wait, status_only=False, dispatcher_addr=dispatcher_addr
        )

        manifest = ResultSchema.model_validate(export["result_export"])

        # sort the nodes
        manifest.lattice.transport_graph.nodes.sort(key=lambda x: x.id)

        rm = ResultManager(manifest, results_dir)
        result_object = rm.result_object
        result_object._results_dir = results_dir
        Path(results_dir).mkdir(parents=True, exist_ok=True)

        # Create node subdirectories
        for node_id in result_object.lattice.transport_graph._graph.nodes:
            node_dir = f"{results_dir}/node_{node_id}"
            Path(node_dir).mkdir(exist_ok=True)

        return rm


def get_result_manager(dispatch_id, results_dir=None, wait=False, dispatcher_addr=None):
    if not results_dir:
        results_dir = get_config("sdk.results_dir") + f"/{dispatch_id}"
    return ResultManager.from_dispatch_id(dispatch_id, results_dir, wait, dispatcher_addr)


def _get_result_multistage(
    dispatch_id: str,
    wait: bool = False,
    dispatcher_addr: str = None,
    status_only: bool = False,
    results_dir: Optional[str] = None,
    *,
    workflow_output: bool = True,
    intermediate_outputs: bool = True,
    sublattice_results: bool = True,
    qelectron_db: bool = False,
) -> Result:
    """
    Get the results of a dispatch from a file.

    Args:
        dispatch_id: The dispatch id of the result.
        wait: Controls how long the method waits for the server to return a result. If False, the method will not wait and will return the current status of the workflow. If True, the method will wait for the result to finish and keep retrying for sys.maxsize.
        status_only: If true, only returns result status, not the full result object. Default is False.
        dispatcher_addr: Dispatcher server address, defaults to the address set in Covalent's config.
        results_dir: The directory where the results are stored in dispatch id named folders.


    Kwargs:
        workflow_output: Whether to return the workflow output. Defaults to True.
        intermediate_outputs: Whether to return all intermediate outputs in the compute graph. Defaults to True.
        sublattice_results: Whether to recursively retrieve sublattice results. Default is True.
        qelectron_db: Whether to load the bytes data of qelectron_db. Default is False.

    Returns:
        The Result object from the Covalent server

    """

    try:
        if status_only:
            return _get_result_export_from_dispatcher(
                dispatch_id=dispatch_id,
                wait=wait,
                status_only=status_only,
                dispatcher_addr=dispatcher_addr,
            )
        rm = get_result_manager(dispatch_id, results_dir, wait, dispatcher_addr)
        _get_default_assets(rm)

        if workflow_output:
            rm.download_result_asset("result")
            rm.load_result_asset("result")

        if intermediate_outputs:
            tg = rm.result_object.lattice.transport_graph
            for node_id in tg._graph.nodes:
                rm.download_node_asset(node_id, "output")
                rm.load_node_asset(node_id, "output")

        if qelectron_db:
            for node_id in tg._graph.nodes:
                rm.download_node_asset(node_id, "qelectron_db")
                rm.load_node_asset(node_id, "qelectron_db")

        # Fetch sublattice result objects recursively
        tg = rm.result_object.lattice.transport_graph
        for node_id in tg._graph.nodes:
            sub_dispatch_id = tg.get_node_value(node_id, "sub_dispatch_id")
            if sublattice_results and sub_dispatch_id:
                sub_result = _get_result_multistage(
                    sub_dispatch_id,
                    wait,
                    dispatcher_addr,
                    status_only,
                    results_dir=results_dir,
                    workflow_output=workflow_output,
                    intermediate_outputs=intermediate_outputs,
                    sublattice_results=sublattice_results,
                    qelectron_db=qelectron_db,
                )
                tg.set_node_value(node_id, "sublattice_result", sub_result)
            else:
                tg.set_node_value(node_id, "sublattice_result", None)

    except MissingLatticeRecordError as ex:
        app_log.warning(
            f"Dispatch ID {dispatch_id} was not found in the database. Incorrect dispatch id."
        )

        raise ex

    return rm.result_object


def get_result(
    dispatch_id: str,
    wait: bool = False,
    dispatcher_addr: str = None,
    status_only: bool = False,
    *,
    results_dir: Optional[str] = None,
    workflow_output: bool = True,
    intermediate_outputs: bool = True,
    sublattice_results: bool = True,
    qelectron_db: bool = False,
) -> Result:
    """
    Get the results of a dispatch.

    Args:
        dispatch_id: The dispatch id of the result.
        wait: Controls how long the method waits for the server to return a result. If False, the method will not wait and will return the current status of the workflow. If True, the method will wait for the result to finish and keep retrying for sys.maxsize.
        dispatcher_addr: Dispatcher server address. Defaults to the address set in Covalent's config.
        status_only: If true, only returns result status, not the full result object. Default is False.

    Kwargs:
        results_dir: The directory where the results are stored in dispatch id named folders.
        workflow_output: Whether to return the workflow output. Defaults to True.
        intermediate_outputs: Whether to return all intermediate outputs in the compute graph. Defaults to True.
        sublattice_results: Whether to recursively retrieve sublattice results. Default is True.
        qelectron_db: Whether to load the bytes data of qelectron_db. Default is False.

    Returns:
        The Result object from the Covalent server

    """
    max_attempts = int(os.getenv("COVALENT_GET_RESULT_RETRIES", 10))
    num_attempts = 0
    while num_attempts < max_attempts:
        try:
            return _get_result_multistage(
                dispatch_id=dispatch_id,
                wait=wait,
                dispatcher_addr=dispatcher_addr,
                status_only=status_only,
                results_dir=results_dir,
                workflow_output=workflow_output,
                intermediate_outputs=intermediate_outputs,
                sublattice_results=sublattice_results,
                qelectron_db=qelectron_db,
            )

        except RecursionError as re:
            app_log.error(re)
            num_attempts += 1
    raise RuntimeError("Timed out waiting for result. Please retry or check dispatch.")



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
from typing import Union

from covalent._results_manager import Result
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._workflow.lattice import Lattice
from covalent._workflow.transport import _TransportGraph

from . import upsert

app_log = logger.app_log


def persist(record: Union[Result, Lattice, _TransportGraph], electron_id: int = None) -> None:
    """Save Result object to a DataStore. Changes are queued until
    committed by the caller.

    Args:
        record: The entity to persist in the DB
        electron_id: (hack) DB-generated id for the parent electron
            if the workflow is actually a subworkflow
    """

    _initialize_results_dir(record)
    app_log.debug(f"Persisting {record}")
    upsert.persist_result(record, electron_id)
    app_log.debug("persist complete")


def _initialize_results_dir(result):
    """Create the results directory."""

    result_folder_path = os.path.join(
        os.environ.get("COVALENT_DATA_DIR") or get_config("dispatcher.results_dir"),
        f"{result.dispatch_id}",
    )
    Path(result_folder_path).mkdir(parents=True, exist_ok=True)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Create custom sentinels and defaults for Covalent"""

import os
from dataclasses import dataclass, field
from typing import Dict

import dask.system

prefix_separator = ":"

parameter_prefix = f"{prefix_separator}parameter{prefix_separator}"
electron_list_prefix = f"{prefix_separator}electron_list{prefix_separator}"
electron_dict_prefix = f"{prefix_separator}electron_dict{prefix_separator}"
subscript_prefix = f"{prefix_separator}subscripted{prefix_separator}"
generator_prefix = f"{prefix_separator}generated{prefix_separator}"
sublattice_prefix = f"{prefix_separator}sublattice{prefix_separator}"
attr_prefix = f"{prefix_separator}attribute{prefix_separator}"
arg_prefix = f"{prefix_separator}arg{prefix_separator}"
postprocess_prefix = f"{prefix_separator}postprocess{prefix_separator}"

WAIT_EDGE_NAME = "!waiting_edge"


def get_default_sdk_config():
    return {
        "config_file": (
            os.environ.get("COVALENT_CONFIG_DIR")
            or os.environ.get("XDG_CONFIG_DIR")
            or (os.environ["HOME"] + "/.config/covalent")
        )
        + "/covalent.conf",
        "log_dir": (
            os.environ.get("COVALENT_LOGDIR")
            or os.path.join(
                os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
                "covalent",
            )
        ),
        "log_level": os.environ.get("LOGLEVEL", "WARNING").lower(),
        "enable_logging": os.environ.get("COVALENT_LOG_TO_FILE", "true").lower(),
        "executor_dir": os.environ.get("COVALENT_EXECUTOR_DIR")
        or (
            (os.environ.get("XDG_CONFIG_DIR") or (os.environ["HOME"] + "/.config"))
            + "/covalent/executor_plugins"
        ),
        "no_cluster": "true" if os.environ.get("COVALENT_DISABLE_DASK") == "1" else "false",
        "exhaustive_postprocess": "false",
        "dispatch_cache_dir": os.environ.get("COVALENT_DISPATCH_CACHE_DIR")
        or (
            (os.environ.get("XDG_CACHE_HOME") or (os.environ["HOME"] + "/.cache"))
            + "/covalent/dispatches"
        ),
        "task_packing": "true" if os.environ.get("COVALENT_ENABLE_TASK_PACKING") else "false",
        "multistage_dispatch": "false"
        if os.environ.get("COVALENT_DISABLE_MULTISTAGE_DISPATCH") == "1"
        else "true",
        "results_dir": os.environ.get(
            "COVALENT_RESULTS_DIR"
        )  # COVALENT_RESULTS_DIR is where the client downloads workflow artifacts during get_result() which is different from COVALENT_DATA_DIR
        or (
            (os.environ.get("XDG_CACHE_HOME") or (os.environ["HOME"] + "/.cache"))
            + "/covalent/results"
        ),
    }


def get_default_dispatcher_config():
    return {
        "address": os.environ.get("COVALENT_DISPATCHER_ADDR", "localhost"),
        "port": int(os.environ.get("COVALENT_SVC_PORT", 48008)),
        "cache_dir": os.environ.get("COVALENT_CACHE_DIR")
        or os.path.join(
            os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
            "covalent",
        ),
        "results_dir": os.environ.get("COVALENT_DATA_DIR")
        or os.path.join(os.environ["HOME"], ".local/share/covalent/data"),
        "log_dir": os.environ.get("COVALENT_LOGDIR")
        or os.path.join(
            os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
            "covalent",
        ),
        "db_path": os.environ.get("COVALENT_DATABASE")
        or (
            (os.environ.get("XDG_DATA_HOME") or (os.environ["HOME"] + "/.local/share"))
            + "/covalent/dispatcher_db.sqlite"
        ),
        "qelectron_db_path": os.environ.get("COVALENT_DATABASE")
        or (
            (os.environ.get("XDG_DATA_HOME") or (os.environ["HOME"] + "/.local/share"))
            + "/covalent/qelectron_db"
        ),
        "heartbeat_interval": os.environ.get("COVALENT_HEARTBEAT_INTERVAL") or 5,
        "heartbeat_file": os.environ.get("COVALENT_HEARTBEAT_FILE")
        or os.path.join(
            (
                os.environ.get("COVALENT_CACHE_DIR")
                or os.path.join(
                    os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
                    "covalent",
                )
            ),
            "heartbeat",
        ),
        "use_async_dispatcher": os.environ.get("COVALENT_USE_ASYNC_DISPATCHER", "true") or "false",
        "data_uri_filter_policy": os.environ.get("COVALENT_DATA_URI_FILTER_POLICY", "http"),
        "asset_cache_size": int(os.environ.get("COVALENT_ASSET_CACHE_SIZE", 32)),
    }


def get_default_dask_config():
    return {
        "cache_dir": os.environ.get("COVALENT_CACHE_DIR")
        or os.path.join(
            os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
            "covalent",
        ),
        "log_dir": os.environ.get("COVALENT_LOGDIR")
        or os.path.join(
            os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
            "covalent",
        ),
        "mem_per_worker": "auto",
        "threads_per_worker": 1,
        "num_workers": dask.system.CPU_COUNT,
    }


def get_default_workflow_data_config():
    return {
        "storage_type": "local",
        "base_dir": os.environ.get("COVALENT_DATA_DIR")
        or (
            (os.environ.get("XDG_DATA_HOME") or (os.environ.get("HOME") + "/.local/share"))
            + "/covalent/workflow_data"
        ),
    }


def get_default_ui_config():
    return {
        "address": "localhost",
        "port": int(os.environ.get("COVALENT_SVC_PORT"))
        if os.environ.get("COVALENT_SVC_PORT")
        else 48008,
        "dev_port": 49009,
        "log_dir": os.environ.get("COVALENT_LOGDIR")
        or os.path.join(
            os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
            "covalent",
        ),
    }


def get_default_executor() -> dict:
    """
    Gets the executor based on whether Dask service is running.

    Returns:
        "dask" as the executor if Dask is running and "local" if Dask is not running.
    """
    from .config import get_config

    return (
        "local"
        if os.environ.get("COVALENT_DISABLE_DASK") == "1" or get_config("sdk.no_cluster") == "true"
        else "dask"
    )


# Default configuration settings
@dataclass
class DefaultConfig:
    sdk: Dict = field(default_factory=get_default_sdk_config)
    dispatcher: Dict = field(default_factory=get_default_dispatcher_config)
    dask: Dict = field(default_factory=get_default_dask_config)
    workflow_data: Dict = field(default_factory=get_default_workflow_data_config)
    user_interface: Dict = field(default_factory=get_default_ui_config)


@dataclass
class DefaultMetadataValues:
    executor: str = field(default_factory=get_default_executor)
    executor_data: Dict = field(default_factory=dict)
    hooks: Dict = field(default_factory=dict)
    workflow_executor: str = field(default_factory=get_default_executor)
    workflow_executor_data: Dict = field(default_factory=dict)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Functions to transform ResultSchema -> Result"""

from datetime import datetime
from typing import Dict, List, Optional

from sqlalchemy.orm import Session

from covalent._shared_files import logger
from covalent._shared_files.schemas.edge import EdgeSchema
from covalent._shared_files.schemas.transport_graph import TransportGraphSchema
from covalent_dispatcher._dal.edge import ElectronDependency
from covalent_dispatcher._dal.electron import Electron
from covalent_dispatcher._dal.job import Job
from covalent_dispatcher._dal.lattice import Lattice
from covalent_dispatcher._db import models
from covalent_dispatcher._object_store.base import BaseProvider

from .electron import import_electron

app_log = logger.app_log


def import_transport_graph(
    session: Session,
    dispatch_id: str,
    tg: TransportGraphSchema,
    lat: Lattice,
    object_store: BaseProvider,
    electron_id: Optional[int],
) -> TransportGraphSchema:
    electron_map = {}
    output_nodes = []

    # Propagate parent electron id's `cancel_requested` property to the sublattice electrons
    if electron_id is not None:
        parent_e_record = Electron.meta_type.get_by_primary_key(session, electron_id)
        job_record = Job.get_by_primary_key(
            session=session, primary_key=parent_e_record.job_id, for_update=True
        )
        cancel_requested = job_record.cancel_requested
    else:
        cancel_requested = False

    # Gather nodes into task groups
    task_groups = {i: [] for i in range(len(tg.nodes))}
    for node in tg.nodes:
        gid = node.metadata.task_group_id
        task_groups[gid].append(node)

    gids = {k: list(map(lambda n: n.id, v)) for k, v in task_groups.items()}

    gid_job_record_map = {}

    # Maps node ids to asset record dictionaries
    electron_asset_links = {}

    for gid in task_groups:
        # Create a job record for each task group
        job_kwargs = {
            "cancel_requested": cancel_requested,
        }

        gid_job_record_map[gid] = Job.create(session, insert_kwargs=job_kwargs, flush=False)

    # Write job records to DB and retrieve primary keys

    session.flush()

    for gid, node_group in task_groups.items():
        for node in node_group:
            job_record = gid_job_record_map[gid]
            e_record, asset_records_by_key, node = import_electron(
                session,
                dispatch_id,
                node,
                lat,
                object_store,
                job_id=job_record.id,
            )
            output_nodes.append(node)
            electron_map[node.id] = e_record
            electron_asset_links[node.id] = asset_records_by_key

    # Compute asset ids, electron ids, and create associations

    n_records = len(electron_map)
    st = datetime.now()
    session.flush()
    et = datetime.now()
    delta = (et - st).total_seconds()
    app_log.debug(f"Inserting {n_records} electron records took {delta} seconds")

    n_records = sum(
        len(asset_records_by_key) for asset_records_by_key in electron_asset_links.values()
    )

    st = datetime.now()
    session.flush()
    et = datetime.now()
    delta = (et - st).total_seconds()
    app_log.debug(f"Inserting {n_records} asset records took {delta} seconds")

    meta_asset_associations = []
    for node_id, asset_records in electron_asset_links.items():
        electron_dal = Electron(session, electron_map[node_id])
        meta_asset_associations.extend(
            electron_dal.associate_asset(session, key, asset_rec.id)
            for key, asset_rec in asset_records.items()
        )
    n_records = len(meta_asset_associations)

    st = datetime.now()
    session.flush()
    et = datetime.now()
    delta = (et - st).total_seconds()
    app_log.debug(f"Inserting {n_records} asset record links took {delta} seconds")

    # Insert edges
    edge_records = []
    edges = [_import_edge(session, e, electron_map, edge_records) for e in tg.links]

    n_records = 0
    n_records = len(edge_records)

    st = datetime.now()
    session.flush()
    et = datetime.now()
    delta = (et - st).total_seconds()
    app_log.debug(f"Inserting {n_records} edge records took {delta} seconds")

    return TransportGraphSchema(nodes=output_nodes, links=edges)


def _import_edge(
    session: Session,
    edge: EdgeSchema,
    electron_map: Dict[int, models.Electron],
    edge_records: List[models.ElectronDependency],
) -> EdgeSchema:
    source_electron = electron_map[edge.source]
    target_electron = electron_map[edge.target]
    edge_name = edge.metadata.edge_name
    param_type = edge.metadata.param_type
    arg_index = edge.metadata.arg_index
    insert_kwargs = {
        "electron_id": target_electron.id,
        "parent_electron_id": source_electron.id,
        "edge_name": edge_name,
        "parameter_type": param_type,
        "arg_index": arg_index,
    }

    edge_records.append(
        ElectronDependency.create(session, insert_kwargs=insert_kwargs, flush=False)
    )

    # No filtering involved
    return edge



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""init

Revision ID: b60c5ecdf927
Revises:
Create Date: 2022-07-19 16:31:18.833092

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
# pragma: allowlist nextline secret
revision = "b60c5ecdf927"
# pragma: allowlist nextline secret
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "electron_dependency",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("electron_id", sa.Integer(), nullable=False),
        sa.Column("parent_electron_id", sa.Integer(), nullable=False),
        sa.Column("edge_name", sa.Text(), nullable=False),
        sa.Column("parameter_type", sa.String(length=24), nullable=True),
        sa.Column("arg_index", sa.Integer(), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=False),
        sa.Column(
            "updated_at",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=True,
        ),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "lattices",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("dispatch_id", sa.String(length=64), nullable=False),
        sa.Column("electron_id", sa.Integer(), nullable=True),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("status", sa.String(length=24), nullable=False),
        sa.Column("electron_num", sa.Integer(), nullable=False),
        sa.Column("completed_electron_num", sa.Integer(), nullable=False),
        sa.Column("storage_type", sa.Text(), nullable=True),
        sa.Column("storage_path", sa.Text(), nullable=True),
        sa.Column("function_filename", sa.Text(), nullable=True),
        sa.Column("function_string_filename", sa.Text(), nullable=True),
        sa.Column("executor_filename", sa.Text(), nullable=True),
        sa.Column("error_filename", sa.Text(), nullable=True),
        sa.Column("inputs_filename", sa.Text(), nullable=True),
        sa.Column("results_filename", sa.Text(), nullable=True),
        sa.Column("transport_graph_filename", sa.Text(), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("started_at", sa.DateTime(), nullable=True),
        sa.Column("completed_at", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "electrons",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("parent_lattice_id", sa.Integer(), nullable=False),
        sa.Column("transport_graph_node_id", sa.Integer(), nullable=False),
        sa.Column("type", sa.String(length=24), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("status", sa.String(length=24), nullable=False),
        sa.Column("storage_type", sa.Text(), nullable=True),
        sa.Column("storage_path", sa.Text(), nullable=True),
        sa.Column("function_filename", sa.Text(), nullable=True),
        sa.Column("function_string_filename", sa.Text(), nullable=True),
        sa.Column("executor_filename", sa.Text(), nullable=True),
        sa.Column("results_filename", sa.Text(), nullable=True),
        sa.Column("value_filename", sa.Text(), nullable=True),
        sa.Column("attribute_name", sa.Text(), nullable=True),
        sa.Column("key_filename", sa.Text(), nullable=True),
        sa.Column("stdout_filename", sa.Text(), nullable=True),
        sa.Column("deps_filename", sa.Text(), nullable=True),
        sa.Column("call_before_filename", sa.Text(), nullable=True),
        sa.Column("call_after_filename", sa.Text(), nullable=True),
        sa.Column("stderr_filename", sa.Text(), nullable=True),
        sa.Column("info_filename", sa.Text(), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("started_at", sa.DateTime(), nullable=True),
        sa.Column("completed_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["parent_lattice_id"],
            ["lattices.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("electrons")
    op.drop_table("lattices")
    op.drop_table("electron_dependency")
    # ### end Alembic commands ###
